/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wfsyn_am.h"

char *progname;

/*
 * Synth NV limits.
 */
#define N_OFF	 64	/* AM "offset" value (0..63) */
#define N_PCT	101	/* AM "percent" value (0..100) */

#ifndef PI
#define PI	3.1415926535897932
#endif /* PI */

static wfsyn_am_map_type new_am_map[N_PCT * N_OFF];
static int n_rows = 0;

/*
 * cmp: qsort-compare function to sort new_am_map rows by power, modulation
 */
static int cmp(const void *vp1, const void *vp2)
{
    const wfsyn_am_map_type *rp1 = vp1;
    const wfsyn_am_map_type *rp2 = vp2;

    if (rp1->m_power > rp2->m_power)
	return  1;
    if (rp1->m_power < rp2->m_power)
	return -1;
    if (rp1->m_depth > rp2->m_depth)
	return  1;
    if (rp1->m_depth < rp2->m_depth)
	return -1;
    return 0;
}

/*
 * vtodbm: convert RMS voltage to dBm at 50 ohms
 */
static double vtodbm(double v)
{
    return 10.0 * log10(v * v / 50.0 * 1000.0);
}

/*
 * dbmtov: convert dBm at 50 ohms to RMS voltage
 */
static double dbmtov(double dbm)
{
    return sqrt(pow(10.0, dbm / 10.0) / 1000.0 * 50.0);
}

/*
 * opxtoa: convert Synth NV offset, percent and instantaneous modulation
 *        voltage normalized to [-1 .. +1] to the Synth NV "amplitude"
 *        value, a.
 */
static int opxtoa(int off, int pct, double x)
{
    return off + (int)rint(pct / 100.0 * (31.5 + 31.999 * x));
}

/*
 * atodbm: convert a Synth NV output "amplitude" value to dBm
 */
static double atodbm(int a)
{
    return 23.0 - (63 - a) / 2.0;
}

/*
 * main
 */
int
main(int argc, char **argv)
{
    int off, pct;
    int i;

    if ((char *)NULL == (progname = strrchr(argv[0], '/'))) {
	    progname = argv[0];
    } else {
	    ++progname;
    }

    /*
     * For each combination of Synth NV offset and percent values...
     */
    for (off = 0; off < N_OFF; ++off) {
	for (pct = 0; pct < N_PCT; ++pct) {
	    wfsyn_am_map_type *mp;
	    double v_min, v_max, v_sumsq, v_rms;
	    
	    /*
	     * If the Synth NV "a" value overflows on a particular
	     * pct value, it will also overflow on higher pct values,
	     * so we can skip to the next offset.
	     */
	    if (opxtoa(off, pct, -1.0) < 0 || opxtoa(off, pct, +1.0) > 63) {
		break;
	    }

	    /*
	     * Init variables.
	     */
	    v_min =  1000;
	    v_max =  -1.0;
	    v_sumsq = 0.0;

	    /*
	     * Compute min, max and rms voltage over one cycle.
	     */
	    for (i = 0; i < WFSYN_N_AM_STEPS; ++i) {
		int a = opxtoa(off, pct, sin(2.0 * PI * i /
					     (double)WFSYN_N_AM_STEPS));
		double v = dbmtov(atodbm(a));

		v_sumsq += v * v;
		if (v < v_min)
		    v_min = v;
		if (v > v_max)
		    v_max = v;
	    }
	    v_rms = sqrt(v_sumsq / (double)WFSYN_N_AM_STEPS);

	    /*
	     * Add a new row to the map.
	     */
	    mp = &new_am_map[n_rows++];
	    mp->m_power = vtodbm(v_rms);
	    mp->m_depth = 100.0 * ((v_max - v_min) / (v_max + v_min));
	    mp->m_off   = off;
	    mp->m_pct   = pct;
	}
    }
    qsort((void *)new_am_map, n_rows, sizeof(wfsyn_am_map_type), cmp);

    /*
     * Print
     */
    (void)printf("/*\n");
    (void)printf(" * This file was generated by %s.  Do not edit.\n", progname);
    (void)printf(" */\n");
    (void)printf("\n");
    (void)printf("#include \"wfsyn_am.h\"\n");
    (void)printf("\n");
    (void)printf("const wfsyn_am_map_type wfsyn_am_map[] = {\n");
    for (i = 0; i < n_rows; ++i) {
	wfsyn_am_map_type *mp = &new_am_map[i];

	(void)printf("    { %6.2f, %6.2f, %3d, %3d }%s\n",
		     mp->m_power, mp->m_depth, mp->m_off, mp->m_pct,
		     i+1 < n_rows ? "," : "");
    }
    (void)printf("};\n");
    (void)printf("\n");
    (void)printf("const int wfsyn_am_map_length = %d;\n", n_rows);

    exit(0);
    /*NOTREACHED*/
}
