/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef WFSYN_H
#define WFSYN_H

#ifndef _INTTYPES_H
#include <inttypes.h>
#endif /* _INTTYPES_H */

#ifndef _MATH_H
#include <math.h>	/* for INFINITY */
#endif /* _MATH_H */

#ifdef __cplusplus
extern "C" {
#endif

typedef unsigned char wfsyn_bool_type;

/*
 * wfsyn_device_type: describes a Synth NV device
 */
struct wfsyn_device;
typedef struct wfsyn_device wfsyn_device_type;

/*
 * WFSYN_MUTE: special output power value used to mute the output
 */
#ifdef INFINITY
#define WFSYN_MUTE	(-INFINITY)
#else /* INFINITY */
#define WFSYN_MUTE	(-1000.0)
#endif /* INFINITY */

/*
 * ADF4350 RF Synthesizer Mode
 */
typedef enum {
    WFSYN_MODE_FRAC_N   = 0,	/* fractional frequency adjustment */
    WFSYN_MODE_LOW_SPUR	= 1,	/* FRAC_N with dither */
    WFSYN_MODE_INT_N    = 2	/* integer frequency adjustment */
} wfsyn_mode_type;

/*
 * Reference Source
 */
typedef enum {
    WFSYN_REFSRC_EXTERNAL = 0,
    WFSYN_REFSRC_INTERNAL = 1
} wfsyn_refsrc_type;

/*
 * ADF4350 Charge Pump Current
 */
typedef enum {
    WFSYN_ICP_0_31MA = 0x0,
    WFSYN_ICP_0_63MA = 0x1,
    WFSYN_ICP_0_94MA = 0x2,
    WFSYN_ICP_1_25MA = 0x3,
    WFSYN_ICP_1_56MA = 0x4,
    WFSYN_ICP_1_88MA = 0x5,
    WFSYN_ICP_2_19MA = 0x6,
    WFSYN_ICP_2_50MA = 0x7,
    WFSYN_ICP_2_81MA = 0x8,
    WFSYN_ICP_3_13MA = 0x9,
    WFSYN_ICP_3_44MA = 0xA,
    WFSYN_ICP_3_75MA = 0xB,
    WFSYN_ICP_4_06MA = 0xC,
    WFSYN_ICP_4_38MA = 0xD,
    WFSYN_ICP_4_69MA = 0xE,
    WFSYN_ICP_5_00MA = 0xF
} wfsyn_icp_type;

/*
 * wfsyn_options_type: configurable parameters
 *
 * Note:
 * 	To maintain binary compatibility, any new options must always be
 *      added to the end of this structure.  Never add/remove members in
 *      the middle.
 */
typedef struct wfsyn_options  {
    double wopt_fref;			/* reference frequency (MHz). */
    double wopt_fres;			/* frequency resolution (MHz). */
    double wopt_fmin;			/* min required output freq. (MHz) */
    double wopt_fmax;			/* max required output freq. (MHz) */
    double wopt_pfdmin;			/* min allowed PFD frequency (MHz) */
    double wopt_pfdmax;			/* max allowed PFD frequency (MHz) */
    wfsyn_mode_type wopt_mode;		/* synthesizer mode */
    wfsyn_refsrc_type wopt_refsrc;	/* reference source */
    wfsyn_bool_type wopt_rmul2en;	/* true to allow ref. doubler. */
    wfsyn_bool_type wopt_wlock;		/* wait for frequency lock */
    wfsyn_icp_type wopt_icp;		/* charge pump current */
} wfsyn_options_type;

/*
 * Step-type values for wfsyn_sweep.
 */
enum {
    WFSYN_STEP_LIN = 0,
    WFSYN_STEP_LOG = 1
};


/*
 * Extended open flags.
 */
enum {
    WFSYN_OPENX_ERROR = 0x00000001
};

/*
 * wfsyn_errfn_type: type of the error handling function passed to openx
 */
typedef void wfsyn_errfn_type(const wfsyn_device_type *wfsynp, const char *msg);

/*
 * Basic Synth Functions
 */
extern wfsyn_device_type *wfsyn_open(const char *device_name);
extern void wfsyn_close(wfsyn_device_type *wfsynp);

extern double wfsyn_get_frequency(const wfsyn_device_type *wfsynp);
extern int wfsyn_set_frequency(wfsyn_device_type *wfsynp, double frequency);

extern double wfsyn_get_power(const wfsyn_device_type *wfsynp);
extern int wfsyn_set_power(wfsyn_device_type *wfsynp, double power);
extern double wfsyn_detect_power(const wfsyn_device_type *wfsynp);

extern int wfsyn_sweep(wfsyn_device_type *wfsynp, double fmin, double fmax,
		       int nsteps, int step_type, double hold_time,
		       const double *calibration_vector,
		       double *frequency_vector, double *result_vector);

/*
 * Modulation Functions
 */
extern int wfsyn_am_burst(wfsyn_device_type *wfsynp,
			  double modulation_frequency, double modulation_depth,
			  unsigned int n_reps);
extern int wfsyn_am_continuous(wfsyn_device_type *wfsynp, wfsyn_bool_type on);

extern int wfsyn_pulse_burst(wfsyn_device_type *wfsynp, double on_time,
			     double off_time, unsigned int n_reps);
extern int wfsyn_pulse_continuous(wfsyn_device_type *wfsynp,
				  wfsyn_bool_type on);

/*
 * Extended Open
 */
extern wfsyn_device_type *wfsyn_openx(const char *device_name,
				      uint32_t openx_flags, ...);

/*
 * Synthesizer Options Control Functions
 */
extern wfsyn_options_type *wfsyn_get_options(wfsyn_device_type *wfsynp);
extern int wfsyn_set_options(wfsyn_device_type *wfsynp);

/*
 * Miscellaneous Functions
 */
extern const char *wfsyn_get_model(const wfsyn_device_type *wfsynp);
extern const char *wfsyn_get_serial(const wfsyn_device_type *wfsynp);
extern const char *wfsyn_get_version(const wfsyn_device_type *wfsynp);
extern void wfsyn_get_rmod(const wfsyn_device_type *wfsynp,
			   wfsyn_bool_type *rmul2p, wfsyn_bool_type *rdiv2p,
			   uint16_t *rp, uint16_t *modp);
extern double wfsyn_get_pfd(const wfsyn_device_type *wfsynp);

/*
 * Output Power Calibration Functions
 */
extern double wfsyn_get_power_offset(const wfsyn_device_type *wfsynp);
extern int wfsyn_set_power_offset(wfsyn_device_type *wfsynp, double offset);

/*
 * Power Detector Calibration Functions
 */
extern double wfsyn_detect_power_raw(const wfsyn_device_type *wfsynp);
extern int wfsyn_set_detector_calibration(wfsyn_device_type *wfsynp, int n,
    					  double (*dp)[2]);

/*
 * Load/Save Functions
 */
extern int wfsyn_load_settings(wfsyn_device_type *wfsynp);
extern int wfsyn_save_settings(const wfsyn_device_type *wfsynp);

extern int wfsyn_save_to_nvram(wfsyn_device_type *wfsynp);

/*
 * General Purpose I/O Functions
 */
extern double wfsyn_read_analog1(const wfsyn_device_type *wfsynp);
extern double wfsyn_read_analog2(const wfsyn_device_type *wfsynp);
extern int wfsyn_write_digital1(const wfsyn_device_type *wfsynp,
				wfsyn_bool_type bit);
extern int wfsyn_write_digital2(const wfsyn_device_type *wfsynp,
				wfsyn_bool_type bit);

#ifdef __cplusplus
}
#endif

#endif /* WFSYN_H */
