/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wfsyn.h"
#include "wfsyn_am.h"
#include "wfsyn_internal.h"

#ifndef PI
#define PI	3.1415926535897932
#endif /* PI */

/*
 * Measured delay between bursts in continuous AM mode.
 */
#define INTER_BURST_DELAY	10.2	/* us */

/*
 * Measured difference between requested step-time and actual step time,
 * i.e. if you ask for 5us, the actual step time is is 3.969.
 */
#define STEP_TIME_OFFSET	1.031	/* us */

/*
 * The V2 firmware limits the maximum step width to 2^14+2 us.
 */
#define STEP_WIDTH_MAX		16386	/* us */

/*
 * Maximum allowed duration of each burst.
 */
#define MAX_BURST		   60	/* s */


/* All table entries plus or minus this amount of power (in dBm)
   are considered when choosing a table entry. */
#define AM_POWER_ERROR	0.5

/*
 * distance: return a "distance" value given power and depth error
 *   @power_error: difference in power (dBm)
 *   @depth_error: difference in depth (percent)
 */
static double distance(double power_error, double depth_error)
{
    /*
     * Scale the inputs so that 0.5 dBm power error is equivalent
     * to 10% AM depth error.
     */
    power_error /= 0.5;
    depth_error /= 10.0;
    return (power_error * power_error + depth_error * depth_error);
}

/*
 * wfsyn_find_am_map_entry: find best AM map entry for the given power & depth
 *   @power: average output power in dBm
 *   @depth: AM depth in percent
 */
static const wfsyn_am_map_type *wfsyn_find_am_map_entry(double power,
							double depth)
{
    int low  = 0;
    int high = wfsyn_am_map_length - 1;
    int cur;
    int best_index;
    double best_distance;

    /*
     * Binary search for an index near the requested power.  At the end of the
     * loop, cur is guaranteed to be within the bounds of the table, and the
     * entry with the closest power is within +/- 1 of cur.
     */
    do {
	cur = (low + high) / 2;
	if (power > wfsyn_am_map[cur].m_power) {
	    low  = cur + 1;
	} else {
	    high = cur - 1;
	}
    } while (low <= high);

    /*
     * Find the table entry with the closest power and set it as the best
     * so far.  This entry may not be within AM_POWER_ERROR of the requested
     * power, but it's the closest possible.
     */
    best_index = cur;
    best_distance = fabs(wfsyn_am_map[cur].m_power - power);
    if (cur - 1 >= 0) {
	double d = fabs(wfsyn_am_map[cur - 1].m_power - power);
	if (d < best_distance) {
	    best_distance = d;
	    best_index = cur - 1;
	}
    }
    if (cur + 1 < wfsyn_am_map_length) {
	double d = fabs(wfsyn_am_map[cur + 1].m_power - power);
	if (d < best_distance) {
	    best_distance = d;
	    best_index = cur + 1;
	}
    }

    /*
     * Now consider all entries within -/+ AM_POWER_ERROR of the
     * requested power value and choose the table entry that
     * minimizes the error.
     */
    low  = best_index;
    high = best_index;
    best_distance = distance(wfsyn_am_map[low].m_power - power,
			     wfsyn_am_map[low].m_depth - depth);
    while (--low >= 0 && wfsyn_am_map[low].m_power >= power - AM_POWER_ERROR) {
	double d = distance(wfsyn_am_map[low].m_power - power,
			    wfsyn_am_map[low].m_depth - depth);
	if (d < best_distance) {
	    best_index = low;
	    best_distance = d;
	}
    }
    while (++high < wfsyn_am_map_length &&
           wfsyn_am_map[high].m_power <= power + AM_POWER_ERROR) {
	double d = distance(wfsyn_am_map[high].m_power - power,
			    wfsyn_am_map[high].m_depth - depth);
	if (d < best_distance) {
	    best_index = high;
	    best_distance = d;
	}
    }
    return &wfsyn_am_map[best_index];
}

/*
 * wfsyn_am_burst: send a burst of n_reps amplitude modulation cycles
 *   @wfsynp: the WFSYN device structure
 *   @modulation_frequency:  AM frequency in Hz
 *   @modulation_depth: AM depth in percent
 *   @n_reps: number of repititions
 */
int
wfsyn_am_burst(wfsyn_device_type *wfsynp, double modulation_frequency,
	       double modulation_depth, unsigned int n_reps)
{
    const wfsyn_am_map_type *mp;
    uint16_t step_time;
    double power;
    double temp;

    /*
     * Validate parameters.
     */
    if (modulation_frequency <= 0) {
	wfsyn_error(wfsynp, "wfsyn_am_burst: %s: %f: "
	                    "invalid modulation_frequency value",
		    wfsynp->wfsyn_device_name, modulation_frequency);
	errno = EINVAL;
	return -1;
    }
    if (modulation_depth < 0.0) {
	wfsyn_error(wfsynp, "wfsyn_am_burst: %s: %f: "
	                    "invalid modulation_depth value",
		    wfsynp->wfsyn_device_name, modulation_depth);
	errno = EINVAL;
	return -1;
    }
    if (n_reps < 1 || n_reps > 65535) {
	wfsyn_error(wfsynp, "wfsyn_am_burst: %s: "
	                    "n_reps value (%u) must be in [1..65535]",
		    wfsynp->wfsyn_device_name, n_reps);
	errno = EINVAL;
	return -1;
    }
    if ((double)n_reps / modulation_frequency > MAX_BURST) {
	wfsyn_error(wfsynp, "wfsyn_am_burst: %s: overall burst duration "
	                    "must be no more than %.0f seconds",
		    wfsynp->wfsyn_device_name, (double)MAX_BURST);
	errno = EINVAL;
	return -1;
    }

    /*
     * If the output is muted, fail.
     */
    if (WFSYN_POWER_IS_MUTED(wfsynp->wfsyn_power)) {
	wfsyn_error(wfsynp, "wfsyn_am_burst: %s: error: output is muted.",
		    wfsynp->wfsyn_device_name);
	errno = EINVAL;
	return -1;
    }

    /*
     * Calculate the step time from the modulation frequency.  We amortize
     * the INTER_BURST_DELAY over the cycle repetitions in a burst so that
     * the behavior in continuous AM mode is more intuitive.
     */
    temp  = 1000000.0 / modulation_frequency;	/* modulation period in us */
    temp -= INTER_BURST_DELAY / (double)n_reps;
    temp /= (double)WFSYN_N_AM_STEPS;		/* convert to step time */
    temp += STEP_TIME_OFFSET;			/* correct for step offset */
    temp  = nearbyint(temp);			/* convert to int */
    if (temp < 1.0) {				/* bound */
	temp = 1.0;
    } else if (temp > STEP_WIDTH_MAX) {
	temp = STEP_WIDTH_MAX;
    }
    step_time = (uint16_t)temp;

    /*
     * If we're in the low power range, add 9 dBm to correct for the fact
     * that the ADF435x will subtract that amount.
     */
    power = wfsynp->wfsyn_power + wfsynp->wfsyn_power_offset;
    if (power <= HMC625_POWER_STEP / 2.0) {
	power += adf435x_power_table[3] - adf435x_power_table[0];
    }
    mp = wfsyn_find_am_map_entry(power, modulation_depth);

    /*
     * Send the command.
     */
    if (wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD) {
	if (step_time != wfsynp->wfsyn_am_step_time) {
	    if (wfsyn_printf(wfsynp, "F%u", step_time) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_am_step_time = step_time;
	}
	if (n_reps != wfsynp->wfsyn_am_reps) {
	    if (wfsyn_printf(wfsynp, "q%u", n_reps) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_am_reps = n_reps;
	}
	if (mp->m_pct != wfsynp->wfsyn_am_pct) {
	    if (wfsyn_printf(wfsynp, "%%%u", mp->m_pct) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_am_pct = mp->m_pct;
	}
	if (mp->m_off != wfsynp->wfsyn_am_off) {
	    if (wfsyn_printf(wfsynp, "@%u", mp->m_off) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_am_off = mp->m_off;
	}
	if (wfsyn_printf(wfsynp, "B") == -1) {
	    return -1;
	}

    } else {
	char buffer[38];
	int i;

	buffer[0] = 9;
	buffer[1] = WFSYN_N_AM_STEPS;
	buffer[2] = (step_time >> 8) & 0xff;
	buffer[3] =  step_time       & 0xff;
	buffer[4] = (n_reps    >> 8) & 0xff;
	buffer[5] =  n_reps          & 0xff;
	for (i = 0; i < WFSYN_N_AM_STEPS; ++i) {
	    temp  = sin(2.0 * PI * (double)i / WFSYN_N_AM_STEPS);
	    temp  = mp->m_pct / 100.0 * (31.5 + 31.999 * temp);
	    temp += mp->m_off;
	    temp  = nearbyint(temp);
	    assert(temp >= 0 && temp <= 63);
	    buffer[6+i] = 0xC0 | (uint8_t)temp;
	}
	if (wfsyn_write(wfsynp, buffer, sizeof(buffer)) == -1) {
	    return -1;
	}
	wfsynp->wfsyn_am_step_time = step_time;
	wfsynp->wfsyn_am_reps = n_reps;
	wfsynp->wfsyn_am_pct = mp->m_pct;
	wfsynp->wfsyn_am_off = mp->m_off;

	/*
	 * Old firmware doesn't restore "a" after an AM burst.  Force
	 * a power update to fix it.
	 */
	wfsynp->wfsyn_hmc625 = 0x40;	/* force power update */
	if (wfsyn_update(wfsynp) == -1) {
	    return -1;
	}
    }
    return 0;
}


/*
 * wfsyn_am_continuous: enable/disable continuous AM
 *   @wfsynp: the WFSYN device structure
 *   @on: 1 to enable; 0 to disable
 */
int
wfsyn_am_continuous(wfsyn_device_type *wfsynp, wfsyn_bool_type on)
{
    /*
     * If the output is muted, don't allow continuous AM on.
     */
    if (WFSYN_POWER_IS_MUTED(wfsynp->wfsyn_power) && on) {
	wfsyn_error(wfsynp, "wfsyn_am_continuous: %s: error: output is muted.",
		    wfsynp->wfsyn_device_name);
	errno = EINVAL;
	return -1;
    }
    if (wfsynp->wfsyn_am_pct == 0xffU) {
	wfsyn_error(wfsynp, "wfsyn_am_continuous: %s: AM parameters not "
		            "initialized; call wfsyn_am_burst first",
			    wfsynp->wfsyn_device_name);
	errno = EINVAL;
	return -1;
    }

    /*
     * Send the command.
     */
    if (wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD) {
	if (wfsyn_printf(wfsynp, "A%u", on ? 1 : 0) == -1) {
	    return -1;
	}
    } else {
	char buffer[1];

	buffer[0] = on ? 11 : 12;
	if (wfsyn_write(wfsynp, buffer, 1) == -1) {
	    return -1;
	}
	/*
	 * Old firmware doesn't restore "a" when continuous AM is turned off.
	 * Force a power update to fix it.
	 */
	if (!on) {
	    wfsynp->wfsyn_hmc625 = 0x40;	/* force power update */
	    if (wfsyn_update(wfsynp) == -1) {
		return -1;
	    }
	}
    }
    return 0;
}
