/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wfsyn.h"
#include "wfsyn_internal.h"

/*
 * wfsyn_set_detector_calibration: set power detector calibration
 *   @wfsynp: the WFSYN device structure
 *   @n: number of rows
 *   @dp: table of x,y pairs, where x represents the raw power detector
 *   	    value and y represents the corresponding power in dBm
 *          The x values must be monotonically increasing.
 */
int
wfsyn_set_detector_calibration(wfsyn_device_type *wfsynp, int n,
			       double (*dp_arg)[2])
{
    double (*dp)[2] = NULL;	/* calibration table */
    double (*scp)[3] = NULL;	/* spline coefficients */

    if (n < 2) {
	wfsyn_error(wfsynp, "%s: wfsyn_set_detector_calibration: %d: "
			    "n must be >= 2",
		    wfsynp->wfsyn_device_name, n);
	errno = EINVAL;
	return -1;
    }
    if ((dp = (double (*)[2])malloc(n * sizeof(double [2]))) == NULL) {
	wfsyn_error(wfsynp, "%s: wfsyn_set_detector_calibration: malloc: %s",
		    wfsynp->wfsyn_device_name, strerror(errno));
	goto error;
    }
    if ((scp = (double (*)[3])malloc((n - 1) * sizeof(double [3]))) == NULL) {
	wfsyn_error(wfsynp, "%s: wfsyn_set_detector_calibration: malloc: %s",
		    wfsynp->wfsyn_device_name, strerror(errno));
	goto error;
    }
    (void)memcpy((void *)dp, (void *)dp_arg, n * sizeof(double [2]));
    if (wfsyn_spline_calc(n - 1, (const double (*)[2])dp, scp) == -1) {
	wfsyn_error(wfsynp, "%s: wfsyn_set_detector_calibration: "
			    "x-coordinates must be monotonically increasing",
		    wfsynp->wfsyn_device_name);
	errno = EINVAL;
	goto error;
    }
    free((void *)wfsynp->wfsyn_dcal_xy);
    wfsynp->wfsyn_dcal_xy = (const double (*)[2])dp;
    free((void *)wfsynp->wfsyn_dcal_bcd);
    wfsynp->wfsyn_dcal_bcd = (const double (*)[3])scp;
    wfsynp->wfsyn_dcal_segments = n - 1;
    return 0;

error:
    if (scp != NULL) {
	free((void *)scp);
	scp = NULL;
    }
    if (dp != NULL) {
	free((void *)dp);
	dp = NULL;
    }
    return -1;
}

/*
 * wfsyn_detect_power_raw: read the raw power detector value
 *   @wfsynp: the WFSYN device structure
 *
 * Returned value is in an unspecified unit; it's not affected by
 * the currently loaded detector calibration curve.
 */
double
wfsyn_detect_power_raw(const wfsyn_device_type *wfsynp)
{
    char buffer[WFSYN_MAX_LINE + 1];

    if (wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD) {
	if (wfsyn_printf(wfsynp, "D") == -1) {
	    return HUGE_VAL;
	}
    } else {
	if (wfsyn_printf(wfsynp, "\004") == -1) {
	    return HUGE_VAL;
	}
    }
    if (wfsyn_readline(wfsynp, buffer, sizeof(buffer)) == -1) {
	return HUGE_VAL;
    }
    return atof(buffer);
}

/*
 * wfsyn_detect_power: read the power detector in dBm
 *   @wfsynp: the WFSYN device structure
 *
 * Returned value is in dBm.
 */
double
wfsyn_detect_power(const wfsyn_device_type *wfsynp)
{
    double x = wfsyn_detect_power_raw(wfsynp);

    if (x == HUGE_VAL) {
	return HUGE_VAL;
    }
    return wfsyn_spline_eval(wfsynp->wfsyn_dcal_segments,
    			     wfsynp->wfsyn_dcal_xy,
			     wfsynp->wfsyn_dcal_bcd,
			     x);
}
