/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <math.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "wfsyn.h"
#include "wfsyn_internal.h"

/*
 * wfsyn_read_analog1: read the J8 P1 analog voltage
 *   @wfsynp:  the WFSYN device structure
 */
double
wfsyn_read_analog1(const wfsyn_device_type *wfsynp)
{
    char buffer[WFSYN_MAX_LINE + 1];

    if (!(wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD)) {
	wfsyn_error(wfsynp, "wfsyn_read_analog1: %s: "
			    "analog read not supported on this device",
		    wfsynp->wfsyn_device_name);
	errno = EPROTONOSUPPORT;
	return HUGE_VAL;
    }
    if (wfsyn_printf(wfsynp, "C1") == -1) {
	return HUGE_VAL;
    }
    if (wfsyn_readline(wfsynp, buffer, sizeof(buffer)) == -1) {
	return HUGE_VAL;
    }
    return 2.5 * (atof(buffer) + 0.5) / 1024.0;
}

/*
 * wfsyn_read_analog2: read the J8 P2 analog voltage
 *   @wfsynp:  the WFSYN device structure
 */
double
wfsyn_read_analog2(const wfsyn_device_type *wfsynp)
{
    char buffer[WFSYN_MAX_LINE + 1];

    if (!(wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD)) {
	wfsyn_error(wfsynp, "wfsyn_read_analog2: %s: "
			    "analog read not supported on this device",
		    wfsynp->wfsyn_device_name);
	errno = EPROTONOSUPPORT;
	return HUGE_VAL;
    }
    if (wfsyn_printf(wfsynp, "C2") == -1) {
	return HUGE_VAL;
    }
    if (wfsyn_readline(wfsynp, buffer, sizeof(buffer)) == -1) {
	return HUGE_VAL;
    }
    return 2.5 * (atof(buffer) + 0.5) / 1024.0;
}

/*
 * wfsyn_write_digital1: write J8 Pin 3
 *   @wfsynp:  the WFSYN device structure
 *   @bit: bit to write
 */
int
wfsyn_write_digital1(const wfsyn_device_type *wfsynp, wfsyn_bool_type bit)
{
    if (!(wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD)) {
	wfsyn_error(wfsynp, "wfsyn_write_digital1: %s: "
			    "digital I/O not supported on this device",
		    wfsynp->wfsyn_device_name);
	errno = EPROTONOSUPPORT;
	return -1;
    }
    return wfsyn_printf(wfsynp, "#%u", bit ? 1 : 0);
}

/*
 * wfsyn_write_digital2: write J8 Pin 5
 *   @wfsynp:  the WFSYN device structure
 *   @bit: bit to write
 */
int
wfsyn_write_digital2(const wfsyn_device_type *wfsynp, wfsyn_bool_type bit)
{
    if (!(wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD)) {
	wfsyn_error(wfsynp, "wfsyn_write_digital2: %s: "
			    "digital I/O not supported on this device",
		    wfsynp->wfsyn_device_name);
	errno = EPROTONOSUPPORT;
	return -1;
    }
    return wfsyn_printf(wfsynp, "$%u", bit ? 1 : 0);
}
