/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <ctype.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wfsyn.h"
#include "wfsyn_internal.h"

/*
 * CRC-8-SAE J1850, polynomial 0x11D table
 */
static uint8_t crc_table[] = {
	0x00, 0x1d, 0x3a, 0x27, 0x74, 0x69, 0x4e, 0x53,
	0xe8, 0xf5, 0xd2, 0xcf, 0x9c, 0x81, 0xa6, 0xbb,
	0xcd, 0xd0, 0xf7, 0xea, 0xb9, 0xa4, 0x83, 0x9e,
	0x25, 0x38, 0x1f, 0x02, 0x51, 0x4c, 0x6b, 0x76,
	0x87, 0x9a, 0xbd, 0xa0, 0xf3, 0xee, 0xc9, 0xd4,
	0x6f, 0x72, 0x55, 0x48, 0x1b, 0x06, 0x21, 0x3c,
	0x4a, 0x57, 0x70, 0x6d, 0x3e, 0x23, 0x04, 0x19,
	0xa2, 0xbf, 0x98, 0x85, 0xd6, 0xcb, 0xec, 0xf1,
	0x13, 0x0e, 0x29, 0x34, 0x67, 0x7a, 0x5d, 0x40,
	0xfb, 0xe6, 0xc1, 0xdc, 0x8f, 0x92, 0xb5, 0xa8,
	0xde, 0xc3, 0xe4, 0xf9, 0xaa, 0xb7, 0x90, 0x8d,
	0x36, 0x2b, 0x0c, 0x11, 0x42, 0x5f, 0x78, 0x65,
	0x94, 0x89, 0xae, 0xb3, 0xe0, 0xfd, 0xda, 0xc7,
	0x7c, 0x61, 0x46, 0x5b, 0x08, 0x15, 0x32, 0x2f,
	0x59, 0x44, 0x63, 0x7e, 0x2d, 0x30, 0x17, 0x0a,
	0xb1, 0xac, 0x8b, 0x96, 0xc5, 0xd8, 0xff, 0xe2,
	0x26, 0x3b, 0x1c, 0x01, 0x52, 0x4f, 0x68, 0x75,
	0xce, 0xd3, 0xf4, 0xe9, 0xba, 0xa7, 0x80, 0x9d,
	0xeb, 0xf6, 0xd1, 0xcc, 0x9f, 0x82, 0xa5, 0xb8,
	0x03, 0x1e, 0x39, 0x24, 0x77, 0x6a, 0x4d, 0x50,
	0xa1, 0xbc, 0x9b, 0x86, 0xd5, 0xc8, 0xef, 0xf2,
	0x49, 0x54, 0x73, 0x6e, 0x3d, 0x20, 0x07, 0x1a,
	0x6c, 0x71, 0x56, 0x4b, 0x18, 0x05, 0x22, 0x3f,
	0x84, 0x99, 0xbe, 0xa3, 0xf0, 0xed, 0xca, 0xd7,
	0x35, 0x28, 0x0f, 0x12, 0x41, 0x5c, 0x7b, 0x66,
	0xdd, 0xc0, 0xe7, 0xfa, 0xa9, 0xb4, 0x93, 0x8e,
	0xf8, 0xe5, 0xc2, 0xdf, 0x8c, 0x91, 0xb6, 0xab,
	0x10, 0x0d, 0x2a, 0x37, 0x64, 0x79, 0x5e, 0x43,
	0xb2, 0xaf, 0x88, 0x95, 0xc6, 0xdb, 0xfc, 0xe1,
	0x5a, 0x47, 0x60, 0x7d, 0x2e, 0x33, 0x14, 0x09,
	0x7f, 0x62, 0x45, 0x58, 0x0b, 0x16, 0x31, 0x2c,
	0x97, 0x8a, 0xad, 0xb0, 0xe3, 0xfe, 0xd9, 0xc4
};

/*
 * Tokens
 */
enum {
    T_FREQUENCY,		/* keyword "frequency" */
    T_POWER,			/* keyword "power" */
    T_POWER_OFFSET,		/* keyword "power_offset" */
    T_OPT_FREF,			/* keyword "opt_fref" */
    T_OPT_FRES,			/* keyword "opt_fres" */
    T_OPT_FMIN,			/* keyword "opt_fmin" */
    T_OPT_FMAX,			/* keyword "opt_fmax" */
    T_OPT_PFDMIN,		/* keyword "opt_pfdmin" */
    T_OPT_PFDMAX,		/* keyword "opt_pfdmax" */
    T_OPT_MODE,			/* keyword "opt_mode" */
    T_OPT_REFSRC,		/* keyword "opt_refsrc" */
    T_OPT_RMUL2EN,		/* keyword "opt_rmul2en" */
    T_OPT_WLOCK,		/* keyword "opt_wlock" */
    T_OPT_ICP,			/* keyword "opt_icp" */
    T_WFSYN_MUTE,		/* keywrod "WFSYN_MUTE" */
    T_DETECTOR_CALIBRATION,	/* keyword "detector_calibration" */
    T_INT,			/* integer constant */
    T_DOUBLE,			/* floating point constant */
    T_ASSIGN,			/* '=' */
    T_LBRACKET,			/* '[' */
    T_RBRACKET,			/* ']' */
    T_NL,			/* ';' or newline */
    T_EOF,			/* end of file */
    T_ERROR			/* lexical error */

};

#define MAX_TOKEN	64

/*
 * Parser State
 */
typedef struct {
    wfsyn_device_type *ps_wfsynp;
    const char *ps_file;
    int ps_line;
    FILE *ps_fp;
    int ps_char;
    int ps_token;
    union {
	int i;
	double lf;
    } ps_value;
    char ps_text[64+1];
} parser_state_type;


#define NEXTC(psp)	((void)((psp)->ps_char = getc((psp)->ps_fp)))

/*
 * skip_to_eol: skip to end of line
 */
static void
skip_to_eol(parser_state_type *psp)
{
    while (psp->ps_char != EOF && psp->ps_char != '\n') {
	NEXTC(psp);
    }
}

/*
 * get_token: return the next token
 */
static void
get_token(parser_state_type *psp)
{
    wfsyn_device_type *wfsynp = psp->ps_wfsynp;
    char *cp = psp->ps_text;
    uint8_t crc = -1;

    *cp = '\000';
    for (;;) {
	/*
	 * In error state, remain in error state.
	 */
	if (psp->ps_token == T_ERROR)
	    return;

	/*
	 * At EOF, keep returning T_EOF.
	 */
	if (psp->ps_char == EOF) {
	    (void)strcpy(psp->ps_text, "<EOF>");
	    psp->ps_token = T_EOF;
	    return;
	}
	/*
	 * If the stream reports an error, report it as EIO.
	 */
	if (ferror(psp->ps_fp) == -1) {
	    wfsyn_error(wfsynp, "%s: %s", psp->ps_file, strerror(errno));
	    (void)strcpy(psp->ps_text, "<ERR>");
	    psp->ps_token = T_ERROR;
	    errno = EIO;
	    return;
	}
	/*
	 * Handle newline.  Skip whitespace.
	 */
	if (isspace(psp->ps_char)) {
	    if (psp->ps_char == '\n') {
		NEXTC(psp);
		++psp->ps_line;
		psp->ps_token = T_NL;
		return;
	    }
	    psp->ps_char = getc(psp->ps_fp);
	    continue;
	}
	/*
	 * Handle keywords.
	 */
	if (isalpha(psp->ps_char)) {
	    do {
		if (cp - psp->ps_text == MAX_TOKEN) {
		    skip_to_eol(psp);
		    *cp = '\000';
		    wfsyn_error(wfsynp, "%s (line %d) error: "
		    		"%s...: keyword too long",
				psp->ps_file, psp->ps_line, psp->ps_text);
		    errno = EINVAL;
		    psp->ps_token = T_ERROR;
		    return;
		}
		*cp++ = psp->ps_char;
		crc = crc_table[crc ^ psp->ps_char];
		NEXTC(psp);
	    } while (isalnum(psp->ps_char) || psp->ps_char == '_');
	    *cp = '\000';
	    //(void)printf("keyword %s CRC 0x%02x\n", psp->ps_text, crc);
	    switch (crc) {
	    case 0x1A:
		if (strcmp(psp->ps_text, "WFSYN_MUTE") == 0) {
		    psp->ps_token = T_WFSYN_MUTE;
		    return;
		}
		break;

	    case 0x35:
		if (strcmp(psp->ps_text, "opt_fres") == 0) {
		    psp->ps_token = T_OPT_FRES;
		    return;
		}
		break;

	    case 0x41:
		if (strcmp(psp->ps_text, "opt_mode") == 0) {
		    psp->ps_token = T_OPT_MODE;
		    return;
		}
		break;

	    case 0x46:
		if (strcmp(psp->ps_text, "detector_calibration") == 0) {
		    psp->ps_token = T_DETECTOR_CALIBRATION;
		    return;
		}
		break;

	    case 0x49:
		if (strcmp(psp->ps_text, "opt_fmax") == 0) {
		    psp->ps_token = T_OPT_FMAX;
		    return;
		}
		break;

	    case 0x51:
		if (strcmp(psp->ps_text, "opt_pfdmin") == 0) {
		    psp->ps_token = T_OPT_PFDMIN;
		    return;
		}
		break;

	    case 0x69:
		if (strcmp(psp->ps_text, "opt_icp") == 0) {
		    psp->ps_token = T_OPT_ICP;
		    return;
		}
		break;

	    case 0x88:
		if (strcmp(psp->ps_text, "opt_pfdmax") == 0) {
		    psp->ps_token = T_OPT_PFDMAX;
		    return;
		}
		break;

	    case 0x90:
		if (strcmp(psp->ps_text, "opt_fmin") == 0) {
		    psp->ps_token = T_OPT_FMIN;
		    return;
		}
		break;

	    case 0x91:
		if (strcmp(psp->ps_text, "opt_fref") == 0) {
		    psp->ps_token = T_OPT_FREF;
		    return;
		}
		break;

	    case 0x97:
		if (strcmp(psp->ps_text, "opt_wlock") == 0) {
		    psp->ps_token = T_OPT_WLOCK;
		    return;
		}
		break;

	    case 0xa1:
		if (strcmp(psp->ps_text, "frequency") == 0) {
		    psp->ps_token = T_FREQUENCY;
		    return;
		}
		break;

	    case 0xb6:
		if (strcmp(psp->ps_text, "power") == 0) {
		    psp->ps_token = T_POWER;
		    return;
		}
		break;

	    case 0xbd:
		if (strcmp(psp->ps_text, "power_offset") == 0) {
		    psp->ps_token = T_POWER_OFFSET;
		    return;
		}
		break;

	    case 0xe4:
		if (strcmp(psp->ps_text, "opt_rmul2en") == 0) {
		    psp->ps_token = T_OPT_RMUL2EN;
		    return;
		}
		break;

	    case 0xee:
		if (strcmp(psp->ps_text, "opt_refsrc") == 0) {
		    psp->ps_token = T_OPT_REFSRC;
		    return;
		}
		break;

	    default:
	    	break;
	    }
	    skip_to_eol(psp);
	    wfsyn_error(wfsynp, "%s (line %d) error: %s: unrecognized keyword",
	    		psp->ps_file, psp->ps_line, psp->ps_text);

	    errno = EINVAL;
	    psp->ps_token = T_ERROR;
	    return;
	}
	/*
	 * Handle numbers.
	 */
	if (isdigit(psp->ps_char) || strchr("+-.", psp->ps_char) != NULL) {
	    char *endp;

	    do {
		if (cp - psp->ps_text == MAX_TOKEN) {
		    skip_to_eol(psp);
		    wfsyn_error(wfsynp, "%s (line %d) error: "
		    		"%s...: number too long",
				psp->ps_file, psp->ps_line, psp->ps_text);
		    errno = EINVAL;
		    psp->ps_token = T_ERROR;
		    return;
		}
		*cp++ = psp->ps_char;
		NEXTC(psp);
	    } while (isdigit(psp->ps_char) ||
	             strchr("+-.e", psp->ps_char) != NULL);
	    *cp = '\000';
	    if (strchr(psp->ps_text, '.') == NULL) {
		psp->ps_value.i = (int)strtol(psp->ps_text, &endp, 0);
		if (endp < cp) {
		    wfsyn_error(wfsynp, "%s (line %d) error: %s: "
		    		"invalid number",
				psp->ps_file, psp->ps_line, psp->ps_text);
		    errno = EINVAL;
		    psp->ps_token = T_ERROR;
		    return;
		}
		psp->ps_token = T_INT;
		return;
	    }
	    psp->ps_value.lf = strtod(psp->ps_text, &endp);
	    if (endp < cp) {
		wfsyn_error(wfsynp, "%s (line %d) error: %s: invalid number",
			    psp->ps_file, psp->ps_line, psp->ps_text);
		errno = EINVAL;
		psp->ps_token = T_ERROR;
		return;
	    }
	    psp->ps_token = T_DOUBLE;
	    return;
	}
	/*
	 * Handle misc tokens.
	 */
	switch (psp->ps_char) {
	case '=':
	    psp->ps_text[0] = psp->ps_char;
	    psp->ps_text[1] = '\000';
	    NEXTC(psp);
	    psp->ps_token = T_ASSIGN;
	    return;
	case ';':
	    psp->ps_text[0] = psp->ps_char;
	    psp->ps_text[1] = '\000';
	    NEXTC(psp);
	    psp->ps_token = T_NL;
	    return;
	case '[':
	    psp->ps_text[0] = psp->ps_char;
	    psp->ps_text[1] = '\000';
	    NEXTC(psp);
	    psp->ps_token = T_LBRACKET;
	    return;
	case ']':
	    psp->ps_text[0] = psp->ps_char;
	    psp->ps_text[1] = '\000';
	    NEXTC(psp);
	    psp->ps_token = T_RBRACKET;
	    return;
	case '#':		/* comment */
	    psp->ps_text[0] = psp->ps_char;
	    psp->ps_text[1] = '\000';
	    skip_to_eol(psp);
	    continue;
	default:
	    break;
	}
	skip_to_eol(psp);
	if (isprint(psp->ps_char)) {
	    wfsyn_error(wfsynp, "%s (line %d) error: invalid character `%c'",
			psp->ps_file, psp->ps_line, psp->ps_char);
	} else {
	    wfsyn_error(wfsynp, "%s (line %d) error: invalid character 0x%02X",
			psp->ps_file, psp->ps_line, psp->ps_char);
	}
	errno = EINVAL;
	psp->ps_token = T_ERROR;
	return;
    }
}

/*
 * parse_double_assignment: parameter _ [=] (double|int) ;
 */
static int
parse_double_assignment(parser_state_type *psp, double *value)
{
    wfsyn_device_type *wfsynp = psp->ps_wfsynp;

    if (psp->ps_token == T_ASSIGN) {
	get_token(psp);
    } else {
	wfsyn_error(wfsynp, "%s (line %d) warning: inserted \"=\"",
		    psp->ps_file, psp->ps_line);
    }
    switch (psp->ps_token) {
    case T_DOUBLE:
	*value = psp->ps_value.lf;
	get_token(psp);
	break;

    case T_INT:
	*value = (double)psp->ps_value.i;
	get_token(psp);
	break;

    case T_WFSYN_MUTE:
	*value = WFSYN_MUTE;
	get_token(psp);
	break;

    case T_ERROR:
	return -1;

    default:
	wfsyn_error(wfsynp, "%s (line %d) error: "
		    "expected floating point value, found \"%s\"",
		    psp->ps_file, psp->ps_line, psp->ps_text);
	errno = EINVAL;
	return -1;
    }
    if (psp->ps_token != T_NL && psp->ps_token != T_EOF) {
	wfsyn_error(wfsynp, "%s (line %d) error: "
		    "unexpected text after assignment: \"%s\"",
		    psp->ps_file, psp->ps_line, psp->ps_text);
	errno = EINVAL;
	return -1;
    }
    get_token(psp);
    return 0;
}

/*
 * parse_int_assignment: parameter _ [=] int ;
 */
static int
parse_int_assignment(parser_state_type *psp, int *value)
{
    wfsyn_device_type *wfsynp = psp->ps_wfsynp;

    if (psp->ps_token == T_ASSIGN) {
	get_token(psp);
    } else {
	wfsyn_error(wfsynp, "%s (line %d) warning: inserted \"=\"",
		    psp->ps_file, psp->ps_line);
    }
    switch (psp->ps_token) {
    case T_INT:
	*value = psp->ps_value.i;
	get_token(psp);
	break;

    case T_ERROR:
	return -1;

    default:
	wfsyn_error(wfsynp, "%s (line %d) error: "
		    "expected integer value; found \"%s\"",
		    psp->ps_file, psp->ps_line, psp->ps_text);
	errno = EINVAL;
	return -1;
    }
    if (psp->ps_token != T_NL && psp->ps_token != T_EOF) {
	wfsyn_error(wfsynp, "%s (line %d) error: "
		    "unexpected text after assignment: \"%s\"",
		    psp->ps_file, psp->ps_line, psp->ps_text);
	errno = EINVAL;
	return -1;
    }
    get_token(psp);
    return 0;
}

/*
 * wfsyn_parse_matrix: parameter = [ _ n,n,n; n,n,n; ... ]
 */
static int
parse_matrix(parser_state_type *psp, int expected_rows, int columns,
	     int *actual_rows, double **vector)
{
    wfsyn_device_type *wfsynp = psp->ps_wfsynp;
    int allocation = 0;
    int index = 0;

    /*
     * Allocate the intial vector.
     */
    allocation = expected_rows * columns;
    if ((*vector = (double *)malloc(allocation * sizeof(double))) == NULL) {
	return -1;
    }

    /*
     * Parse until we hit a right bracket.
     */
    *actual_rows = 0;
    while (psp->ps_token != T_RBRACKET) {
	int current_column = 0;

	if (psp->ps_token == T_NL) {
	    get_token(psp);
	    continue;
	}
	if (psp->ps_token == T_ERROR)
	    return -1;

	if (psp->ps_token != T_DOUBLE && psp->ps_token != T_INT) {
	    wfsyn_error(wfsynp, "%s (line %d) error: "
	    		"expected floating point value; found \"%s\"",
		        psp->ps_file, psp->ps_line, psp->ps_text);
	    errno = EINVAL;
	    return -1;
	}
	current_column = 0;
	do {
	    if (index >= allocation) {
		allocation += (allocation >> 1) + 1;
		if ((*vector =
		     (double *)realloc(*vector, allocation * sizeof(double))
		   ) == NULL) {
		    free((void *)*vector);
		    *vector = NULL;
		    return -1;
		}
	    }
	    if (psp->ps_token == T_DOUBLE) {
		(*vector)[index++] = psp->ps_value.lf;
	    } else {
		(*vector)[index++] = (double)psp->ps_value.i;
	    }
	    ++current_column;
	    get_token(psp);
	} while (psp->ps_token == T_DOUBLE || psp->ps_token == T_INT);

	/*
	 * Handle errors.
	 */
	if (psp->ps_token == T_ERROR) {
	    return -1;
	}
	if (current_column != columns) {
	    wfsyn_error(wfsynp, "%s (line %d) error: expected %d column%s "
	    		        "but only found %d\n",
		        psp->ps_file, psp->ps_line, columns,
			columns != 1 ? "s" : "", current_column);
	    (void)free((void *)*vector);
	    *vector = NULL;
	    return -1;
	}
	if (psp->ps_token != T_NL && psp->ps_token != T_RBRACKET) {
	    wfsyn_error(wfsynp, "%s (line %d) error: "
			        "unexpected text \"%s\" at end of line",
				psp->ps_file, psp->ps_line, psp->ps_text);
	    (void)free((void *)*vector);
	    *vector = NULL;
	    return -1;
	}
	++*actual_rows;
    }
    get_token(psp);
    return 0;
}

/*
 * wfsyn_load_settings: load startup settings
 */
int
wfsyn_load_settings(wfsyn_device_type *wfsynp)
{
    wfsyn_options_type *woptp = &wfsynp->wfsyn_new_options;
    parser_state_type ps;
    int rv = 0;
    int i_temp;
    double lf_temp;

    /*
     * Init parser state.
     */
    (void)memset((void *)&ps, 0, sizeof(parser_state_type));
    ps.ps_wfsynp = wfsynp;
    ps.ps_file = wfsynp->wfsyn_settings_filename;
    ps.ps_line = 1;

    /*
     * Open the file.  Because the startup file is optional, don't
     * print errors on ENOENT.
     */
    if ((ps.ps_fp = fopen(ps.ps_file, "r")) == NULL) {
	if (errno != ENOENT) {
	    wfsyn_error(wfsynp, "%s: %s", ps.ps_file, strerror(errno));
	}
	rv = -1;
	goto out;
    }
    /*
     * Parse the file.
     */
    NEXTC(&ps);
    get_token(&ps);
    for (;;) {
	switch (ps.ps_token) {
	case T_FREQUENCY:
	    get_token(&ps);
	    if (parse_double_assignment(&ps, &lf_temp) == -1) {
		rv = -1;
		break;
	    }
	    wfsynp->wfsyn_frequency = lf_temp;
	    continue;

	case T_POWER:
	    get_token(&ps);
	    if (parse_double_assignment(&ps, &lf_temp) == -1) {
		rv = -1;
		break;
	    }
	    wfsynp->wfsyn_power = lf_temp;
	    continue;

	case T_POWER_OFFSET:
	    get_token(&ps);
	    if (parse_double_assignment(&ps, &lf_temp) == -1) {
		rv = -1;
		break;
	    }
	    wfsynp->wfsyn_power_offset = lf_temp;
	    continue;

	case T_OPT_FREF:
	    get_token(&ps);
	    if (parse_double_assignment(&ps, &lf_temp) == -1) {
		rv = -1;
		break;
	    }
	    woptp->wopt_fref = lf_temp;
	    continue;

	case T_OPT_FRES:
	    get_token(&ps);
	    if (parse_double_assignment(&ps, &lf_temp) == -1) {
		rv = -1;
		break;
	    }
	    woptp->wopt_fres = lf_temp;
	    continue;

	case T_OPT_FMIN:
	    get_token(&ps);
	    if (parse_double_assignment(&ps, &lf_temp) == -1) {
		rv = -1;
		break;
	    }
	    woptp->wopt_fmin = lf_temp;
	    continue;

	case T_OPT_FMAX:
	    get_token(&ps);
	    if (parse_double_assignment(&ps, &lf_temp) == -1) {
		rv = -1;
		break;
	    }
	    woptp->wopt_fmax = lf_temp;
	    continue;

	case T_OPT_PFDMIN:
	    get_token(&ps);
	    if (parse_double_assignment(&ps, &lf_temp) == -1) {
		rv = -1;
		break;
	    }
	    woptp->wopt_pfdmin = lf_temp;
	    continue;

	case T_OPT_PFDMAX:
	    get_token(&ps);
	    if (parse_double_assignment(&ps, &lf_temp) == -1) {
		rv = -1;
		break;
	    }
	    woptp->wopt_pfdmax = lf_temp;
	    continue;

	case T_OPT_MODE:
	    get_token(&ps);
	    if (parse_int_assignment(&ps, &i_temp) == -1) {
		rv = -1;
		break;
	    }
	    if (i_temp != WFSYN_MODE_FRAC_N &&
	        i_temp != WFSYN_MODE_LOW_SPUR &&
	        i_temp != WFSYN_MODE_INT_N) {
		wfsyn_error(wfsynp, "%s (line %d) error: invalid mode value",
				    ps.ps_file, ps.ps_line);
		rv = -1;
		break;
	    }
	    woptp->wopt_mode = i_temp;
	    continue;

	case T_OPT_REFSRC:
	    get_token(&ps);
	    if (parse_int_assignment(&ps, &i_temp) == -1) {
		rv = -1;
		break;
	    }
	    woptp->wopt_refsrc = (i_temp != 0);
	    continue;

	case T_OPT_RMUL2EN:
	    get_token(&ps);
	    if (parse_int_assignment(&ps, &i_temp) == -1) {
		rv = -1;
		break;
	    }
	    woptp->wopt_rmul2en = (i_temp != 0);
	    continue;

	case T_OPT_WLOCK:
	    get_token(&ps);
	    if (parse_int_assignment(&ps, &i_temp) == -1) {
		rv = -1;
		break;
	    }
	    woptp->wopt_wlock = (i_temp != 0);
	    continue;

	case T_OPT_ICP:
	    get_token(&ps);
	    if (parse_int_assignment(&ps, &i_temp) == -1) {
		rv = -1;
		break;
	    }
	    woptp->wopt_icp = (i_temp &
	    		      (ADF435X_R2_ICP_MASK >> ADF435X_R2_ICP_SHIFT));
	    continue;

	case T_DETECTOR_CALIBRATION:
	    {
		int actual_rows = 0;
		double *vector = NULL;
		double (*dp)[2];
		int first_line = ps.ps_line;

		get_token(&ps);
		if (ps.ps_token == T_ASSIGN) {
		    get_token(&ps);
		} else {
		    wfsyn_error(wfsynp, "%s (line %d) warning: inserted \"=\"",
				ps.ps_file, ps.ps_line);
		}
		if (ps.ps_token != T_LBRACKET) {
		    wfsyn_error(wfsynp, "%s (line %d) error: expected \"[\"; "
		    		        "found \"%s\"",
				ps.ps_file, ps.ps_line, ps.ps_text);
		    rv = -1;
		    break;
		}
		get_token(&ps);
		if (parse_matrix(&ps, 17, 2, &actual_rows, &vector) == -1) {
		    rv = -1;
		    break;
		}
		dp = (double (*)[2])vector;
		if (actual_rows < 2) {
		    rv = -1;
		    break;
		}
		if (wfsyn_set_detector_calibration(wfsynp, actual_rows,
						   dp) == -1) {
		    if (errno == EINVAL) {
			wfsyn_error(wfsynp, "%s (line %d) error: "
					    "detector_calibration matrix "
					    "must have at least two rows",
					    ps.ps_file, first_line);
			free((void *)vector);
			rv = -1;
			break;
		    }
		    wfsyn_error(wfsynp, "%s (line %d) error: %s",
		    		ps.ps_file, first_line, strerror(errno));
		    free((void *)vector);
		    rv = -1;
		    break;
		}
		free((void *)vector);
		continue;
	    }

	case T_NL:
	    get_token(&ps);
	    continue;

	case T_EOF:
	    break;

	case T_ERROR:
	    rv = -1;
	    break;

	default:
	    wfsyn_error(wfsynp, "%s (line %d) error: invalid syntax",
			ps.ps_file, ps.ps_line);
	    errno = EINVAL;
	    rv = -1;
	    break;
	}
	break;
    }
out:
    if (ps.ps_fp != NULL) {
	(void)fclose(ps.ps_fp);
	ps.ps_fp = NULL;
    }
    return rv;
}

/*
 * wfsyn_save_settings: save all settings to the start-up file
 */
int
wfsyn_save_settings(const wfsyn_device_type *wfsynp)
{
    const wfsyn_options_type *woptp = &wfsynp->wfsyn_new_options;
    FILE *fp;
    char *cp;
    int i;

    /*
     * Make the directory, if needed.
     */
    if ((cp = strrchr(wfsynp->wfsyn_settings_filename, '/')) != NULL) {
	*cp = '\000';
	if (mkdir(wfsynp->wfsyn_settings_filename, 0777) == -1 &&
	    errno != EEXIST) {
	    wfsyn_error(wfsynp, "error: %s: %s",
			wfsynp->wfsyn_settings_filename, strerror(errno));
	    *cp = '/';
	    return -1;
	}
	*cp = '/';
    }

    /*
     * Write out the new settings.
     *
     * TODO - it would be better to write to a new file and then use atomic
     * rename to replace the old so that we don't clobber the existing file on
     * error.
     */
    if ((fp = fopen(wfsynp->wfsyn_settings_filename, "w")) == NULL) {
	wfsyn_error(wfsynp, "%s: %s", wfsynp->wfsyn_settings_filename,
		    strerror(errno));
	return 1;
    }
    (void)fprintf(fp, "frequency    = %11.6f # MHz\n",
		  wfsynp->wfsyn_frequency);
    if (WFSYN_POWER_IS_MUTED(wfsynp->wfsyn_power)) {
	(void)fprintf(fp, "power        =  WFSYN_MUTE # dBm\n");
    } else {
	(void)fprintf(fp, "power        =  %5.1f      # dBm\n",
		      wfsynp->wfsyn_power);
    }
    (void)fprintf(fp, "power_offset =  %5.1f      # dBm\n",
                  wfsynp->wfsyn_power_offset);
    (void)fprintf(fp, "opt_fmin     = %11.6f # MHz\n",
    		  woptp->wopt_fmin);
    (void)fprintf(fp, "opt_fmax     = %11.6f # MHz\n",
    		  woptp->wopt_fmax);
    (void)fprintf(fp, "opt_fref     = %11.6f # MHz\n",
    		  woptp->wopt_fref);
    (void)fprintf(fp, "opt_fres     = %11.6f # MHz\n",
    		  woptp->wopt_fres);
    (void)fprintf(fp, "opt_icp      =   %2d\n",
    		  woptp->wopt_icp);
    (void)fprintf(fp, "opt_mode     =    %1d\n",
    		  (int)woptp->wopt_mode);
    (void)fprintf(fp, "opt_pfdmin   = %11.6f # MHz\n",
    		  woptp->wopt_pfdmin);
    (void)fprintf(fp, "opt_pfdmax   = %11.6f # MHz\n",
    		  woptp->wopt_pfdmax);
    (void)fprintf(fp, "opt_refsrc   =    %1d\n",
    		  (int)woptp->wopt_refsrc);
    (void)fprintf(fp, "opt_rmul2en  =    %1d\n",
    		  (int)woptp->wopt_rmul2en);
    (void)fprintf(fp, "opt_wlock    =    %1d\n",
    		  (int)woptp->wopt_wlock);
    (void)fprintf(fp, "\n");
    (void)fprintf(fp, "detector_calibration = [\n");
    for (i = 0; i <= wfsynp->wfsyn_dcal_segments; ++i) {
	(void)fprintf(fp, "    %10.6f %10.6f\n",
		      wfsynp->wfsyn_dcal_xy[i][0],
		      wfsynp->wfsyn_dcal_xy[i][1]);
    }
    (void)fprintf(fp, "]\n");

    if (fclose(fp) == -1) {
	return -1;
    }
    return 0;
}
