/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>
#include "wfsyn.h"
#include "wfsyn_internal.h"

/*
 * get_startup_filename: compute wfsyn_settings_filename
 */
static int
get_settings_filename(wfsyn_device_type *wfsynp)
{
    const char *model  = wfsyn_get_model(wfsynp);
    const char *serial = wfsyn_get_serial(wfsynp);
    const char *wfsyn_dir = NULL;
    const char *home_dir = NULL;
    char *filename = NULL;
    char *base;
    char *cur;
    char *cp;
    size_t size = 0;
    static const char dot_dir[] = ".wfsyn";

    /*
     * Calculate the string length.  If there is no serial number,
     * use the device basename;
     */
    if ((wfsyn_dir = getenv("WFSYN_DIR")) != NULL) {
	size = strlen(wfsyn_dir) + 1;
    } else if ((home_dir = getenv("HOME")) != NULL) {
	size = strlen(home_dir) + 1 + sizeof(dot_dir);
    }
    size += strlen(model) + 1;
    if (serial != NULL) {
	size += strlen(serial) + 1;

    } else {
	if ((cp = strrchr(wfsynp->wfsyn_device_name, '/')) == NULL) {
	    serial = wfsynp->wfsyn_device_name;
	} else {
	    serial = cp + 1;
	}
	size += strlen(serial) + 1;
    }

    /*
     * Allocate a buffer for the filename.
     */
    if ((filename = malloc(size)) == NULL) {
	return -1;
    }
    cur = filename;

    /*
     * Copy the directory name.
     */
    if (wfsyn_dir != NULL) {
	cur += sprintf(cur, "%s/", wfsyn_dir);
    } else if (home_dir != NULL) {
	cur += sprintf(cur, "%s/%s/", home_dir, dot_dir);
    }

    /*
     * Copy the basename.
     */
    base = cur;
    cur += sprintf(cur, "%s:%s", model, serial);
    assert(cur - filename + 1 == size);

    /*
     * Change whitespace to underscore, upper-case to lower-case, and
     * slashes to vertical lines.
     */
    for (cp = base; *cp != '\000'; ++cp) {
	if (*cp == ' ' || *cp == '\t') {
	    *cp = '_';
	} else if (isupper(*cp)) {
	    *cp = tolower(*cp);
	} else if (*cp == '/') {
	    *cp = '|';
	}
    }
    wfsynp->wfsyn_settings_filename = filename;
    return 0;
}

/*
 * wfsyn_dcal_default: default power detector calibration curve
 *    y = 0.117748 * x - 85.0515
 */
static double wfsyn_dcal_default[][2] = {
    { 301.0, -49.60935  },
    { 722.0,  -0.037444 },
};

/*
 * wfsyn_open: open a SynthNV device
 *   @device_name: name of WFSYN TTY device
 */
wfsyn_device_type *
wfsyn_open(const char *device_name)
{
    return wfsyn_openx(device_name, 0U);
}

/*
 * wfsyn_openx: open a SynthNV device
 *   @device_name: name of WFSYN TTY device
 *   @flags: extended open flags
 */
wfsyn_device_type *
wfsyn_openx(const char *device_name, uint32_t flags, ...)
{
    va_list ap;
    char const * function_name = ((flags == 0) ? "wfsyn_open" : "wfsyn_openx");
    wfsyn_errfn_type *errfn = wfsyn_default_error_function;
    wfsyn_device_type *wfsynp;
    struct termios termios;
    int rc;

    /*
     * Handle extended arguments.
     */
    va_start(ap, flags);
    if (flags & WFSYN_OPENX_ERROR) {
	errfn = va_arg(ap, wfsyn_errfn_type *);
    }
    va_end(ap);

    /*
     * Allocate and zero the device structure.
     */
    if ((wfsynp = malloc(sizeof(wfsyn_device_type))) == NULL) {
	wfsyn_error(NULL, "%s: %s: malloc %s",
		    function_name, device_name, strerror(errno));
	goto error;
    }
    (void)memset((void *)wfsynp, 0, sizeof(wfsyn_device_type));
    wfsynp->wfsyn_errfn = errfn;
    wfsynp->wfsyn_fd = -1;

    /*
     * Save the device name.
     */
    if ((wfsynp->wfsyn_device_name = malloc(strlen(device_name) + 1)) == NULL) {
	goto error;
    }
    (void)strcpy(wfsynp->wfsyn_device_name, device_name);

    /*
     * Open the device and set raw mode.
     */
    if ((wfsynp->wfsyn_fd = open(device_name, O_RDWR|O_NONBLOCK, 0)) == -1) {
	wfsyn_error(wfsynp, "%s: %s: %s",
		    function_name, device_name, strerror(errno));
	goto error_free_device;
    }
    if (tcgetattr(wfsynp->wfsyn_fd, &termios) == -1) {
	wfsyn_error(wfsynp, "%s: tcgetattr: %s: %s",
		    function_name, device_name, strerror(errno));
	goto error_free_device;
    }
    termios.c_iflag |=	IGNBRK;
    termios.c_iflag &= ~BRKINT;
    termios.c_iflag |=	IGNPAR;
    termios.c_iflag &= ~PARMRK;
    termios.c_iflag &= ~INPCK;
    termios.c_iflag &= ~ISTRIP;
    termios.c_iflag &= ~INLCR;
    termios.c_iflag &= ~IGNCR;
    termios.c_iflag &= ~ICRNL;
#ifdef IUCLC
    termios.c_iflag &= ~IUCLC;
#endif /* IUCLC */
    termios.c_iflag &= ~IXON;
    termios.c_iflag &= ~IXANY;
    termios.c_iflag &= ~IXOFF;
    termios.c_iflag &= ~IMAXBEL;
#ifdef IUTF8
    termios.c_iflag &= ~IUTF8;
#endif /* IUTF8 */
    termios.c_oflag &= ~OPOST;
#ifdef OLCUC
    termios.c_oflag &= ~OLCUC;
#endif /* OLCUC */
    termios.c_oflag &= ~ONLCR;
    termios.c_oflag &= ~OCRNL;
    termios.c_oflag &= ~ONOCR;
    termios.c_oflag &= ~ONLRET;
    termios.c_oflag &= ~OFILL;
    termios.c_cflag &= ~CSIZE;
    termios.c_cflag |=	CS8;
    termios.c_cflag |=	CREAD;
    termios.c_cflag &= ~PARENB;
    termios.c_lflag &= ~ISIG;
    termios.c_lflag &= ~ICANON;
    termios.c_lflag &= ~ECHO;
#ifdef ECHOCTL
    termios.c_lflag &= ~ECHOCTL;
#endif /* ECHCTL */
    termios.c_lflag &= ~TOSTOP;
#ifdef PENDIN
    termios.c_lflag &= ~PENDIN;
#endif /* PENDIN */
    if (tcsetattr(wfsynp->wfsyn_fd, TCSANOW, &termios) == -1) {
	goto error_free_device;
    }

    /*
     * Set the close on exec flag.
     */
    if ((rc = fcntl(wfsynp->wfsyn_fd, F_GETFD)) != -1) {
	rc |= FD_CLOEXEC;
	(void)fcntl(wfsynp->wfsyn_fd, F_SETFD);
    }

    /*
     * Turn off O_NONBLOCK.
     */
    if ((rc = fcntl(wfsynp->wfsyn_fd, F_GETFL)) == -1) {
	wfsyn_error(wfsynp, "%s: fcntl (F_GETFL): %s: %s",
	            function_name, device_name, strerror(errno));
	goto error_free_device;
    }
    rc &= ~O_NONBLOCK;
    if (fcntl(wfsynp->wfsyn_fd, F_SETFL, (long)rc) == -1) {
	wfsyn_error(wfsynp, "%s: fcntl (F_SETFL): %s: %s",
	            function_name, device_name, strerror(errno));
	goto error_free_device;
    }

    /*
     * Read the version string and determine capabilities.
     */
    if (wfsyn_read_version(wfsynp) == -1)
	goto error_free_device;

    /*
     * Get the startup filename.
     */
    if (get_settings_filename(wfsynp) == -1)
	goto error_free_device;

    /*
     * Set initial values.
     */
    wfsynp->wfsyn_frequency = 1000.0; /* 1GHz */
    wfsynp->wfsyn_power = -10.0; /* dBm */
    wfsynp->wfsyn_current_options.wopt_fref =	10.0; /* MHz */
    wfsynp->wfsyn_current_options.wopt_fres =	 1.0; /* MHz */
    wfsynp->wfsyn_current_options.wopt_fmin =
	(wfsynp->wfsyn_capabilities & WFSYN_CAP_ADF4351) ?
	    34.375 : 137.5; /* MHz */
    wfsynp->wfsyn_current_options.wopt_fmax = 4400.0; /* MHz */
    wfsynp->wfsyn_current_options.wopt_pfdmin =	 1.0; /* MHz */
    wfsynp->wfsyn_current_options.wopt_pfdmax =
	(wfsynp->wfsyn_capabilities & WFSYN_CAP_ADF4351) ?
	    ADF4351_PFD_MAX : ADF4350_PFD_MAX;
    wfsynp->wfsyn_current_options.wopt_refsrc  = WFSYN_REFSRC_INTERNAL;
    wfsynp->wfsyn_current_options.wopt_rmul2en = 1;
    wfsynp->wfsyn_current_options.wopt_wlock =
	(wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD) ? 1 : 0;
    wfsynp->wfsyn_current_options.wopt_mode = WFSYN_MODE_FRAC_N;
    wfsynp->wfsyn_current_options.wopt_icp = WFSYN_ICP_5_00MA;
    wfsynp->wfsyn_am_pct = 0xFF;
    wfsynp->wfsyn_am_off = 0xFF;

    /*
     * Set a default power detector calibration curve.
     */
    if (wfsyn_set_detector_calibration(wfsynp,
				       sizeof(wfsyn_dcal_default) /
					 sizeof(double [2]),
				       wfsyn_dcal_default) == -1) {
	goto error_free_device;
    }

    /*
     * Modify the above with values read from the device, if supported.
     */
    (void)wfsyn_read_registers(wfsynp);

    /*
     * Modify the above with values read from the startup file, if it exists.
     */
    if (wfsyn_load_settings(wfsynp) == -1 && errno != ENOENT) {
	goto error_free_device;
    }

    /*
     * Compute PFD and update the device.
     */
    wfsynp->wfsyn_new_options = wfsynp->wfsyn_current_options;
    if (wfsyn_set_options(wfsynp) == -1) {
	goto error_free_device;
    }

    /*
     * Read the power detector and discard the result.  This is a
     * work-around for a firmware problem where the first call to
     * detect_power returns an unreasonably low value.
     */
    (void)wfsyn_detect_power_raw(wfsynp);

    return wfsynp;

error_free_device:
    if (wfsynp != NULL) {
	wfsyn_close(wfsynp);
    }
error:
    return NULL;
}

/*
 * wfsyn_close: close a SynthNV device
 *   @wfsynp: the WFSYN device structure
 */
void
wfsyn_close(wfsyn_device_type *wfsynp)
{
    if (wfsynp->wfsyn_fd != -1) {
	(void)close(wfsynp->wfsyn_fd);
	wfsynp->wfsyn_fd = -1;
    }
    if (wfsynp->wfsyn_dcal_bcd != NULL) {
	free((void *)wfsynp->wfsyn_dcal_bcd);
	wfsynp->wfsyn_dcal_bcd = NULL;
    }
    if (wfsynp->wfsyn_dcal_xy != NULL) {
	free((void *)wfsynp->wfsyn_dcal_xy);
	wfsynp->wfsyn_dcal_xy = NULL;
    }
    if (wfsynp->wfsyn_settings_filename != NULL) {
	free((void *)wfsynp->wfsyn_settings_filename);
	wfsynp->wfsyn_settings_filename = NULL;
    }
    if (wfsynp->wfsyn_serial_string != NULL) {
	free((void *)wfsynp->wfsyn_serial_string);
	wfsynp->wfsyn_serial_string = NULL;
    }
    if (wfsynp->wfsyn_model_string != NULL) {
	free((void *)wfsynp->wfsyn_model_string);
	wfsynp->wfsyn_model_string = NULL;
    }
    if (wfsynp->wfsyn_version_string != NULL) {
	free((void *)wfsynp->wfsyn_version_string);
	wfsynp->wfsyn_version_string = NULL;
    }
    free((void *)wfsynp);
}

/*
 * wfsyn_get_model: return the hardware model string
 *   @wfsynp: the WFSYN device structure
 */
const char *
wfsyn_get_model(const wfsyn_device_type *wfsynp)
{
    /*
     * Firmware version 1 doesn't support the model string.  Assume
     * this is firmware is only seen on hardware version 1.
     */
    if (wfsynp->wfsyn_model_string == NULL) {
	return "WFT SynthNV 1";
    }
    return wfsynp->wfsyn_model_string;
}

/*
 * wfsyn_get_version: return the firmware version string
 *   @wfsynp: the WFSYN device structure
 */
const char *
wfsyn_get_version(const wfsyn_device_type *wfsynp)
{
    return wfsynp->wfsyn_version_string;
}

/*
 * wfsyn_get_serial: return the serial number string (or NULL if none)
 *   @wfsynp: the WFSYN device structure
 */
const char *
wfsyn_get_serial(const wfsyn_device_type *wfsynp)
{
    return wfsynp->wfsyn_serial_string;
}
