/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wfsyn.h"
#include "wfsyn_internal.h"

/*
 * Measured difference between requested pulse width and actual pulse
 * width, i.e. if you ask for an on-time of 5us, the actual width is 3.58us.
 */
#define PULSE_ON_OFFSET		1.42	/* us */
#define PULSE_OFF_OFFSET	1.08	/* us */


/*
 * The V2 firmware limits the maximum on/off wtime with to 2^14+2 us.
 */
#define PULSE_MAX	16386	/* us */

/*
 * Maximum allowed duration of each burst.
 */
#define MAX_BURST		   60	/* s */

/*
 * wfsyn_pulse_burst: send a burst of n_reps pulses
 *   @wfsynp: the WFSYN device structure
 *   @on_time:  pulse on  time in us
 *   @off_time: pulse off time in us
 *   @n_reps: number of repititions
 */
int
wfsyn_pulse_burst(wfsyn_device_type *wfsynp, double on_time, double off_time,
		  unsigned int n_reps)
{
    uint16_t on_us;
    uint16_t off_us;

    /*
     * Validate the on/off time parameters.
     */
    if (on_time < 0.0 || on_time > PULSE_MAX) {
	wfsyn_error(wfsynp, "wfsyn_pulse_burst: %s: "
	                    "on_time value (%g) must be less than %u us",
		    wfsynp->wfsyn_device_name, on_time, PULSE_MAX);
	errno = EINVAL;
	return -1;
    }
    if (off_time < 0.0 || off_time > PULSE_MAX) {
	wfsyn_error(wfsynp, "wfsyn_pulse_burst: %s: "
	                    "off_time value (%g) must be less than %u us",
		    wfsynp->wfsyn_device_name, off_time, PULSE_MAX);
	errno = EINVAL;
	return -1;
    }
    if (n_reps < 1 || n_reps > 65535) {
	wfsyn_error(wfsynp, "wfsyn_pulse_burst: %s: "
	                    "n_reps value (%u) must be in [1..65535]",
		    wfsynp->wfsyn_device_name, n_reps);
	errno = EINVAL;
	return -1;
    }
    if ((on_time + off_time) * n_reps / 1000000.0 > MAX_BURST) {
	wfsyn_error(wfsynp, "wfsyn_pulse_burst: %s: overall burst "
	                    "duration must be less than %.0f seconds",
		    wfsynp->wfsyn_device_name, (double)MAX_BURST);
	errno = EINVAL;
	return -1;
    }

    /*
     * Add the offset time t on_time, bound and convert to uint16_t.
     */
    on_time  = nearbyint(on_time  + PULSE_ON_OFFSET);
    if (on_time < 1.0) {
	on_time = 1.0;
    } else if (on_time > PULSE_MAX) {
	on_time = PULSE_MAX;
    }
    on_us = (uint16_t)on_time;

    /*
     * Add the offset time t off_time, bound and convert to uint16_t.
     */
    off_time = nearbyint(off_time + PULSE_OFF_OFFSET);
    if (off_time < 1.0) {
	off_time = 1.0;
    } else if (off_time > PULSE_MAX) {
	off_time = PULSE_MAX;
    }
    off_us = (uint16_t)off_time;

    /*
     * Send the command.
     */
    if (wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD) {
	if (on_us != wfsynp->wfsyn_pulse_on) {
	    if (wfsyn_printf(wfsynp, "P%u", on_us) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_pulse_on = on_us;
	}
	if (off_us != wfsynp->wfsyn_pulse_off) {
	    if (wfsyn_printf(wfsynp, "O%u", off_us) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_pulse_off = off_us;
	}
	if (n_reps != wfsynp->wfsyn_pulse_reps) {
	    if (wfsyn_printf(wfsynp, "R%u", n_reps) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_pulse_reps = n_reps;
	}
	if (wfsyn_printf(wfsynp, "G") == -1) {
	    return -1;
	}

    } else {
	char buffer[7];

	buffer[0] = 10;
	buffer[1] = (on_us  >>  8) & 0xff;
	buffer[2] =  on_us         & 0xff;
	buffer[3] = (off_us >>  8) & 0xff;
	buffer[4] =  off_us        & 0xff;
	buffer[5] = (n_reps >>  8) & 0xff;
	buffer[6] =  n_reps        & 0xff;
	if (wfsyn_write(wfsynp, buffer, 7) == -1) {
	    return -1;
	}
	wfsynp->wfsyn_pulse_on   = on_us;
	wfsynp->wfsyn_pulse_off  = off_us;
	wfsynp->wfsyn_pulse_reps = n_reps;
    }
    wfsynp->wfsyn_hmc625 = 0x40;
    return 0;
}

/*
 * wfsyn_pulse_continuous: enable/disable continuous pulse burst
 *   @wfsynp: the WFSYN device structure
 *   @on: 1 to enable; 0 to disable
 */
int
wfsyn_pulse_continuous(wfsyn_device_type *wfsynp, wfsyn_bool_type on)
{
    if (wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD) {
	if (wfsyn_printf(wfsynp, "j%u", on ? 1 : 0) == -1) {
	    return -1;
	}
    } else {
	char buffer[1];

	buffer[0] = on ? 13 : 14;
	if (wfsyn_write(wfsynp, buffer, 1) == -1) {
	    return -1;
	}
    }
    wfsynp->wfsyn_hmc625 = 0x40;
    return 0;
}
