/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wfsyn.h"
#include "wfsyn_internal.h"


/*
 * wfsyn_read_registers: read the current settings if possible
 *   @wfsynp: the WFSYN device structure
 */
int
wfsyn_read_registers(wfsyn_device_type *wfsynp)
{
    wfsyn_options_type *woptp = &wfsynp->wfsyn_current_options;
    char buffer[WFSYN_MAX_LINE + 1];
    int i;
    int rfdiv;
    double f_ratio;	/* frequency / pfd ratio */
    double r_ratio;	/* pfd / reference ratio */
    double temp;
    double V, Q;
    double dcal[2][2];

    if (wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD) {
	/*
	 * Read the ADF435x registers.
	 */
	for (i = 0; i < 6; ++i) {
	    if (wfsyn_printf(wfsynp, "H%d?", i) == -1) {
		return -1;
	    }
	    if (wfsyn_readline(wfsynp, buffer, sizeof(buffer)) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_adf435x[i] = strtol(buffer, NULL, 16);
	}

	/*
	 * Read the HMC625 value.
	 */
	if (wfsyn_printf(wfsynp, "a?") == -1) {
	    return -1;
	}
	if (wfsyn_readline(wfsynp, buffer, sizeof(buffer)) == -1) {
	    return -1;
	}
	wfsynp->wfsyn_hmc625 = atoi(buffer) | 0xC0;

	/*
	 * Get the current frequency and bound it to the device limits.
	 */
	if (wfsyn_printf(wfsynp, "f?") == -1) {
	    return -1;
	}
	if (wfsyn_readline(wfsynp, buffer, sizeof(buffer)) == -1) {
	    return -1;
	}
	wfsynp->wfsyn_frequency = atof(buffer) / 1000.0;
	if (wfsynp->wfsyn_frequency < woptp->wopt_fmin) {
	    wfsynp->wfsyn_frequency = woptp->wopt_fmin;
	} else if (wfsynp->wfsyn_frequency > woptp->wopt_fmax) {
	    wfsynp->wfsyn_frequency = woptp->wopt_fmax;
	}

	/*
	 * Read the internal reference flag.
	 */
	if (wfsyn_printf(wfsynp, "x?") == -1) {
	    return -1;
	}
	if (wfsyn_readline(wfsynp, buffer, sizeof(buffer)) == -1) {
	    return -1;
	}
	wfsynp->wfsyn_refsrc = (atoi(buffer) != 0);

	/*
	 * Calculate the mode.
	 */
	if (wfsynp->wfsyn_adf435x[2] & ADF435X_R2_LDF) {
		woptp->wopt_mode = WFSYN_MODE_INT_N;
	} else if (((wfsynp->wfsyn_adf435x[2] & ADF435X_R2_LNLS_MASK) >>
		   ADF435X_R2_LNLS_SHIFT) == 3) {
		woptp->wopt_mode = WFSYN_MODE_LOW_SPUR;
	} else {
		woptp->wopt_mode = WFSYN_MODE_FRAC_N;
	}

	/*
	 * Calculate the output frequency to PFD ratio.
	 */
	rfdiv = (wfsynp->wfsyn_adf435x[4] & ADF435X_R4_RF_DIVIDER_MASK) >>
		ADF435X_R4_RF_DIVIDER_SHIFT;
	f_ratio = (((wfsynp->wfsyn_adf435x[0] & ADF435X_R0_INT_MASK) >>
		    ADF435X_R0_INT_SHIFT) +
		   ((wfsynp->wfsyn_adf435x[0] & ADF435X_R0_FRAC_MASK) >>
		    ADF435X_R0_FRAC_SHIFT) / (double)
		   ((wfsynp->wfsyn_adf435x[1] & ADF435X_R1_MOD_MASK) >>
		    ADF435X_R1_MOD_SHIFT)) /
	          (double)(1U << rfdiv);

	/*
	 * Calculate the PFD to reference frequency ratio.
	 */
	r_ratio = (double)((wfsynp->wfsyn_adf435x[2] & ADF435X_R2_R_MASK) >>
		           ADF435X_R2_R_SHIFT);
	if (wfsynp->wfsyn_adf435x[2] & ADF435X_R2_RMUL2)
		r_ratio *= 2.0;
	if (wfsynp->wfsyn_adf435x[2] & ADF435X_R2_RDIV2)
		r_ratio /= 2.0;

	/*
	 * Deduce the reference frequency from the "f" frequency and
	 * the ratios computed from the registers.
	 */
	woptp->wopt_fref = wfsynp->wfsyn_frequency / f_ratio / r_ratio;

	/*
	 * If the internal reference is enabled and the reference
	 * frequency isn't with 5% of 10 MHz, then assume the frequency
	 * setting is wrong and set it to 10 MHz.
	 */
	if (wfsynp->wfsyn_refsrc &&
	    (woptp->wopt_fref < 9.5 || woptp->wopt_fref > 10.5)) {
	    woptp->wopt_fref = 10.0;
	}

	/*
	 * Calculate the coarsest frequency resolution assuming an RF
	 * divider value of 1.
	 */
	temp = woptp->wopt_fref * r_ratio;	/* PFD frequency */
	if (woptp->wopt_mode != WFSYN_MODE_INT_N) {
		temp /= (wfsynp->wfsyn_adf435x[1] & ADF435X_R1_MOD_MASK) >>
			ADF435X_R1_MOD_SHIFT;
	}
	woptp->wopt_fres = temp;

	/*
	 * Calculate the output power.
	 */
	if ((wfsynp->wfsyn_adf435x[4] & ADF435X_R4_RFOUT_ENABLE) &&
	    (wfsynp->wfsyn_hmc625 & HMC625_ENABLE)) {
	    wfsynp->wfsyn_power = adf435x_power_table[
		(wfsynp->wfsyn_adf435x[4] & ADF435X_R4_RFOUT_POWER_MASK) >>
		    ADF435X_R4_RFOUT_POWER_SHIFT] +
		HMC625_POWER0 +
		    (double)(wfsynp->wfsyn_hmc625 & HMC625_POWER_MASK) *
		    HMC625_POWER_STEP;
	} else {
	    wfsynp->wfsyn_power = WFSYN_MUTE;
	}

	/*
	 * Read the detector reference voltage.
	 */
	if (wfsyn_printf(wfsynp, "V?") == -1) {
	    return -1;
	}
	if (wfsyn_readline(wfsynp, buffer, sizeof(buffer)) == -1) {
	    return -1;
	}
	V = atof(buffer);

	/*
	 * Read the detector offset value.
	 */
	if (wfsyn_printf(wfsynp, "Q?") == -1) {
	    return -1;
	}
	if (wfsyn_readline(wfsynp, buffer, sizeof(buffer)) == -1) {
	    return -1;
	}
	Q = atof(buffer);

	/*
	 * Override the default detector calibration curve with the
	 * slope just read.  The formula is:
	 *
	 *          x   *   V               1023 * 0.020 * (y + Q)
	 *   y = --------------- - Q,  x = ------------------------
	 *        1023  * 0.020                       V
	 *
	 * We give two arbitrary points: y = -50 dBm and y = 0 dBm.  With
	 * two points, the spline interpolator makes a straight line through
	 * the points.  If V and Q are totally unreasonable (such that the
	 * slope is negative), keep the default settings and continue.
	 */
	dcal[0][1] = -50.0; /* y = -50.0 dBm */
	dcal[0][0] = 1023.0 * 0.020 * (dcal[0][1] + Q) / V;
	dcal[1][1] =   0.0; /* y =   0.0 dBm */
	dcal[1][0] = 1023.0 * 0.020 * (dcal[1][1] + Q) / V;
        if (wfsyn_set_detector_calibration(wfsynp, 2, dcal) == -1 &&
	    errno != EINVAL) {
		return -1;
	}

	/*
	 * Get the charge pump current from R2.
	 */
	woptp->wopt_icp = (wfsynp->wfsyn_adf435x[2] & ADF435X_R2_ICP_MASK) >>
			  ADF435X_R2_ICP_SHIFT;

    }
    return 0;
}
