/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <fcntl.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "wfsyn.h"
#include "wfsyn_internal.h"


/*
 * wfsyn_read_version: get the firmware version, capabilities, model and serial
 *   @wfsynp: the WFSYN device structure
 */
int
wfsyn_read_version(wfsyn_device_type *wfsynp)
{
    int rv;
    char *cp;
    char *cp_end;
    char buffer[WFSYN_MAX_LINE + 1];
    unsigned int v1, v2;

    /*
     * Send a control-G to request the version string.
     */
    if (wfsyn_printf(wfsynp, "\007") == -1) {
	return -1;
    }

    /*
     * Read the version string.	 Firmware version 1 didn't send the
     * terminating newline, and so has to be special-cased.
     */
    cp_end = buffer;
    for (;;) {
	int remaining = &buffer[WFSYN_MAX_LINE] - cp_end;

	if (remaining == 0) {
	    return -1;
	}
	if ((rv = read(wfsynp->wfsyn_fd, cp_end, remaining)) == -1) {
	    wfsyn_error(wfsynp, "read: %s: %s",
			wfsynp->wfsyn_device_name, strerror(errno));
	    return -1;
	}
	if (rv == 0) {
	    wfsyn_error(wfsynp, "read: %s: unexpected EOF",
	    		wfsynp->wfsyn_device_name);
	    errno = EIO;
	    return -1;
	}
	*(cp_end += rv) = '\000';
	if (cp_end[-1] == '\n') {
	    *--cp_end = '\000';
	    if (cp_end > buffer && cp_end[-1] == '\r') {
		*--cp_end = '\000';
	    }
	    break;
	}
	if ((cp = strrchr(buffer, 'V')) != NULL) {
	    if (sscanf(cp, "Version %u.%u", &v1, &v2) == 2 &&
		v1 == 1)
		break;
	}
    }
    /*
     * Copy the version string.
     */
    if (wfsynp->wfsyn_version_string != NULL)
    {
	free((void *)wfsynp->wfsyn_version_string);
	wfsynp->wfsyn_version_string = NULL;
    }
    if ((wfsynp->wfsyn_version_string = malloc(cp_end - buffer + 1)) == NULL) {
	return -1;
    }
    (void)memcpy((void *)wfsynp->wfsyn_version_string, (void *)buffer,
	         cp_end - buffer + 1);

    /*
     * Parse the version string and determine capabilities.
     */
    if (sscanf(buffer, "WFT SynthNV Version %u.%u",
	   &v1, &v2) == 2) {
	goto success;
    }
    if (sscanf(buffer, "WFT SynthNV Firmware Version %u.%u",
	   &v1, &v2) == 2) {
	goto success;
    }
    wfsyn_error(wfsynp, "wfsyn_read_version: %s: %s: "
    		"unsupported firmware version",
    		wfsynp->wfsyn_device_name, wfsynp->wfsyn_version_string);
    errno = EPROTONOSUPPORT;
    return -1;

success:
    /*
     * Beginning with firmware version 2, the commands change from binary to
     * text format, the generator is based on the ADF4351 chip and the read
     * MUX command is supported.
     */
    if (v1 >= 2) {
	wfsynp->wfsyn_capabilities |= WFSYN_CAP_TEXTCMD |
				      WFSYN_CAP_ADF4351 |
				      WFSYN_CAP_READMUX;
    }

    /*
     * If the device has model and serial number support, read and
     * copy the model and serial numbers.
     */
    if (wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD)
    {
	size_t length;

	/*
	 * Get model.
	 */
	if (wfsynp->wfsyn_model_string != NULL) {
		free((void *)wfsynp->wfsyn_model_string);
		wfsynp->wfsyn_model_string = NULL;
	}
	if (wfsyn_printf(wfsynp, "+") == -1) {
		return -1;
	}
	if (wfsyn_readline(wfsynp, buffer, sizeof(buffer)) == -1) {
		return -1;
	}
	length = strlen(buffer);
	if (length > 0) {
	    if ((wfsynp->wfsyn_model_string = malloc(length + 1)) == NULL) {
		return -1;
	    }
	    (void)memcpy(wfsynp->wfsyn_model_string, buffer, length + 1);
	}

	/*
	 * Get serial.
	 */
	if (wfsynp->wfsyn_serial_string != NULL) {
		free((void *)wfsynp->wfsyn_serial_string);
		wfsynp->wfsyn_serial_string = NULL;
	}
	if (wfsyn_printf(wfsynp, "-") == -1) {
		return -1;
	}
	if (wfsyn_readline(wfsynp, buffer, sizeof(buffer)) == -1) {
		return -1;
	}
	length = strlen(buffer);
	if (length > 0) {
	    if ((wfsynp->wfsyn_serial_string = malloc(length + 1)) == NULL) {
		return -1;
	    }
	    (void)memcpy(wfsynp->wfsyn_serial_string, buffer, length + 1);
	}
    }
    return 0;
}
