/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wfsyn.h"
#include "wfsyn_internal.h"

/*
 * Table of prime numbers less than 1024.
 */
static uint16_t prime_table[] = {
       2,    3,	   5,	 7,   11,   13,	  17,	19,
      23,   29,	  31,	37,   41,   43,	  47,	53,
      59,   61,	  67,	71,   73,   79,	  83,	89,
      97,  101,	 103,  107,  109,  113,	 127,  131,
     137,  139,	 149,  151,  157,  163,	 167,  173,
     179,  181,	 191,  193,  197,  199,	 211,  223,
     227,  229,	 233,  239,  241,  251,	 257,  263,
     269,  271,	 277,  281,  283,  293,	 307,  311,
     313,  317,	 331,  337,  347,  349,	 353,  359,
     367,  373,	 379,  383,  389,  397,	 401,  409,
     419,  421,	 431,  433,  439,  443,	 449,  457,
     461,  463,	 467,  479,  487,  491,	 499,  503,
     509,  521,	 523,  541,  547,  557,	 563,  569,
     571,  577,	 587,  593,  599,  601,	 607,  613,
     617,  619,	 631,  641,  643,  647,	 653,  659,
     661,  673,	 677,  683,  691,  701,	 709,  719,
     727,  733,	 739,  743,  751,  757,	 761,  769,
     773,  787,	 797,  809,  811,  821,	 823,  827,
     829,  839,	 853,  857,  859,  863,	 877,  881,
     883,  887,	 907,  911,  919,  929,	 937,  941,
     947,  953,	 967,  971,  977,  983,	 991,  997,
    1009, 1013, 1019, 1021
};
#define N_PRIMES    (sizeof(prime_table) / sizeof(uint16_t))


/*
 * wfsyn_get_options: return a pointer to an initialized options struct
 *   @wfsynp: the WFSYN device structure
 *
 * The caller may modify the returned struct and call wfsyn_set_options
 * to put the new options into effect.	The memory for returned struct
 * is part of the wfsyn_device_type and the caller should not call free
 * on the returned pointer.
 */
wfsyn_options_type *
wfsyn_get_options(wfsyn_device_type *wfsynp)
{
    wfsynp->wfsyn_new_options = wfsynp->wfsyn_current_options;
    return &wfsynp->wfsyn_new_options;
}

/*
 * wfsyn_set_options: put new options into effect
 *   @wfsynp: the WFSYN device structure
 */
int
wfsyn_set_options(wfsyn_device_type *wfsynp)
{
    const wfsyn_options_type *woptp = &wfsynp->wfsyn_new_options;
    wfsyn_bool_type rmul2_possible;
    wfsyn_bool_type pre45_possible;
    wfsyn_bool_type rmul2_enable;
    wfsyn_bool_type rdiv2_enable;
    uint16_t int_min;
    uint32_t r = 1;
    uint32_t mod;
    uint32_t mod_min;
    uint32_t mod_max;
    int rfdiv_min = wfsyn_calc_rfdiv(wfsynp, woptp->wopt_fmin);
    int rfdiv_max = wfsyn_calc_rfdiv(wfsynp, woptp->wopt_fmax);
    int prime_index = 0;
    double vco_fmax45;
    double fref2;
    double temp;
    double vco_fmin;
    double pfd_max;

    /*
     * Validate parameters.
     */
    if (woptp->wopt_fref <= 0.0) {
	wfsyn_error(wfsynp, "wfsyn_set_options: %s: %f: "
	                    "invalid wopt_fref value",
		    wfsynp->wfsyn_device_name, woptp->wopt_fref);
	errno = EINVAL;
	return -1;
    }
    if (woptp->wopt_fres <= 0.0) {
	wfsyn_error(wfsynp, "wfsyn_set_options: %s: %f: "
	                    "invalid wopt_fres value",
		    wfsynp->wfsyn_device_name, woptp->wopt_fres);
	errno = EINVAL;
	return -1;
    }
    if (woptp->wopt_fmin <= 0.0) {
	wfsyn_error(wfsynp, "wfsyn_set_options: %s: %f:"
	                    "invalid wopt_fmin value",
		    wfsynp->wfsyn_device_name, woptp->wopt_fmin);
	errno = EINVAL;
	return -1;
    }
    if (woptp->wopt_fmin > woptp->wopt_fmax) {
	wfsyn_error(wfsynp, "wfsyn_set_options: %s: "
	                    "wopt_fmin (%f) > wopt_fmax (%f)",
		    wfsynp->wfsyn_device_name,
		    woptp->wopt_fmin, woptp->wopt_fmax);
	errno = EINVAL;
	return -1;
    }

    if (woptp->wopt_pfdmin <= 0.0) {
	wfsyn_error(wfsynp, "wfsyn_set_options: %s: %f: "
	                    "invalid wopt_pfdmin value",
		    wfsynp->wfsyn_device_name, woptp->wopt_pfdmin);
	errno = EINVAL;
	return -1;
    }
    if (woptp->wopt_pfdmin > woptp->wopt_pfdmax) {
	wfsyn_error(wfsynp, "wfsyn_set_options: %s: "
	                    "wopt_pfdmin (%f) > wopt_pfdmax (%f)",
		    wfsynp->wfsyn_device_name,
		    woptp->wopt_pfdmin, woptp->wopt_pfdmax);
	errno = EINVAL;
	return -1;
    }
    if (woptp->wopt_refsrc & ~0x1U) {
	wfsyn_error(wfsynp, "wfsyn_set_options: %s: %u:"
	                    "invalid wopt_refsrc value",
		    wfsynp->wfsyn_device_name, woptp->wopt_refsrc);
	errno = EINVAL;
	return -1;
    }
    if (woptp->wopt_icp < WFSYN_ICP_0_31MA ||
        woptp->wopt_icp > WFSYN_ICP_5_00MA) {
	wfsyn_error(wfsynp, "wfsyn_set_options: %s: %u:"
	                    "invalid wopt_icp value",
		    wfsynp->wfsyn_device_name, woptp->wopt_icp);
	errno = EINVAL;
	return -1;
    }

    /*
     * Make sure wlock is not set on the revision 1 models.
     */
    if (woptp->wopt_wlock &&
        !(wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD)) {
	wfsyn_error(wfsynp, "wfsyn_set_options: %s: "
			    "wopt_wlock not supported on this device",
		    wfsynp->wfsyn_device_name);
	errno = EPROTONOSUPPORT;
	return -1;
    }

    /*
     * Calculate the maximum allowed PFD frequency.
     */
    if ((wfsynp->wfsyn_capabilities & WFSYN_CAP_ADF4351) &&
	 woptp->wopt_mode != WFSYN_MODE_INT_N) {
	pfd_max = MIN(woptp->wopt_pfdmax, ADF4351_PFD_MAX);
    } else {
	pfd_max = MIN(woptp->wopt_pfdmax, ADF4350_PFD_MAX);
    }

    /*
     * Determine if the reference doubler can be used.
     */
    if (woptp->wopt_rmul2en && woptp->wopt_fref <= ADF435X_RMUL2_FMAX) {
	rmul2_possible = 1;
    } else {
	rmul2_possible = 0;
    }

    /*
     * Determine if the 4/5 prescalar can be used.  This is possible only when
     * the VCO frequency remains below ADF435x_45FMAX over the entire frequency
     * range.  The prescaler determines the minimum VCO frequency and minimum
     * INT value.
     */
    vco_fmax45 = (wfsynp->wfsyn_capabilities & WFSYN_CAP_ADF4351) ?
	    ADF4351_45FMAX : ADF4350_45FMAX;
    if (rfdiv_min == rfdiv_max &&
	woptp->wopt_fmax * (double)(1 << rfdiv_max) <= vco_fmax45) {
	vco_fmin = woptp->wopt_fmin * (double)(1 << rfdiv_max);
	int_min = ADF434X_INT_MIN_PRE45;
	pre45_possible = 1;
    } else {
	vco_fmin = ADF435X_VCO_MIN;
	int_min = ADF434X_INT_MIN_PRE89;
	pre45_possible = 0;
    }

    /*
     * Compute min and max modulus values based on mode.  The final min modulus
     * value programmed is always 2, but internally, we can use 1.
     */
    switch (woptp->wopt_mode) {
    case WFSYN_MODE_FRAC_N:
	mod_min = 1;
	mod_max = ADF435X_MOD_MAX;
	break;
    case WFSYN_MODE_LOW_SPUR:
	mod_min = ADF435X_MOD_MIN_LS;
	mod_max = ADF435X_MOD_MAX;
	break;
    case WFSYN_MODE_INT_N:
	mod_min = 1;
	mod_max = 1;
	break;
    default:
	wfsyn_error(wfsynp, "wfsyn_set_options: %s: %u:"
	                    "invalid wopt_mode value",
		    wfsynp->wfsyn_device_name, woptp->wopt_mode);
	errno = EINVAL;
	return -1;
    }

    /*
     * Strategy 1: try to find an R value consisting of the smallest prime
     * factors that can be divided out of mod.  Begin by computing the mod
     * value for R=1, then move factors from mod to R until mod is in
     * range.
     */
    rmul2_enable = rmul2_possible;
    rdiv2_enable = 0;
    fref2 = woptp->wopt_fref * (rmul2_enable ? 2.0 : 1.0);
    mod = MAX(1, (uint32_t)nearbyint(fref2 / (woptp->wopt_fres *
		                     (double)(1 << rfdiv_max))));

    /*
     * Divide prime factors out of mod and multiply R while mod is too large,
     * PFD is too high, or INT is too low.
     */
    while (mod > mod_max ||
	   fref2 / (double)r > pfd_max ||
	   floor(vco_fmin / (fref2 / (double)r)) < int_min) {

	/*
	 * If no prime factors remain, fail.
	 */
	if (prime_index >= N_PRIMES ||
	    mod < prime_table[prime_index]) {
	    goto strategy2;
	}
	/*
	 * If R would become too large, fail.
	 */
	if (r * prime_table[prime_index] > ADF435X_R_MAX) {
	    goto strategy2;
	}
	/*
	 * If the PFD frequency would become too small, fail.
	 */
	if (fref2 / (double)r / (double)prime_table[prime_index] <
	    woptp->wopt_pfdmin) {
	    goto strategy2;
	}
	/*
	 * Test if mod is divisible by the current prime.  If it is,
	 * divide the current prime out.
	 */
	if (mod % prime_table[prime_index] == 0) {
	    mod /= prime_table[prime_index];

	    /*
	     * If we're dividing out a factor of 2, first try disabling
	     * the reference doubler, next try enabling the reference
	     * divider.  If neither of those is available or we're
	     * diving something other than 2, then multiply R.
	     */
	    if (prime_index == 0 && rmul2_enable) {
		rmul2_enable = 0;
		fref2 = woptp->wopt_fref;
	    } else if (prime_index == 0 && !rdiv2_enable) {
		rdiv2_enable = 1;
		fref2 = woptp->wopt_fref / 2.0;
	    } else {
		r *= prime_table[prime_index];
	    }
	    continue;
	}
	++prime_index;
    }

    /*
     * If the reference divider is enabled, but R can take another factor
     * of two, disable the reference divider and double R.
     */
    if (rdiv2_enable && r < ADF435X_R_MAX / 2)
    {
	rdiv2_enable = 0;
	fref2 = woptp->wopt_fref;
	r <<= 1;
    }
    goto success;

strategy2:
    /*
     * Strategy2: find the minimum R that puts mod in range.
     */
    rmul2_enable = rmul2_possible;
    rdiv2_enable = 0;
    fref2 = woptp->wopt_fref * (rmul2_enable ? 2.0 : 1.0);

    /*
     * Find the smallest value of R that provides the frequested frequency
     * resolution with mod <= mod_max, makes pfd <= pfdmax, and that makes
     * INT >= int_min.
     */
    temp = fref2 / woptp->wopt_fres * (double)(1 << rfdiv_max) /
           (double)mod_max;
    temp = MAX(temp, fref2 / woptp->wopt_pfdmax);
    temp = MAX(temp, (double)int_min / (vco_fmin / fref2));
    r = (uint32_t)ceil(temp);

    /*
     * If the reference multiplier is enabled and R is either too large
     * or is even, disable the reference multiplier.
     */
    if ((r > ADF435X_R_MAX || (r & 1) == 0) && rmul2_enable) {
	rmul2_enable = 0;
	fref2 = woptp->wopt_fref;
	r >>= 1;
    }
    /*
     * If R is still too large, enable the reference divider.
     */
    if (r > ADF435X_R_MAX) {
	rdiv2_enable = 1;
	fref2 = woptp->wopt_fref / 2.0;
	r >>= 1;
    }
    /*
     * If R is still too large or if the PFD frequency is too low, fail.
     */
    if (r > ADF435X_R_MAX) {
	goto fail;
    }
    if (fref2 / (double)r < woptp->wopt_pfdmin) {
	goto fail;
    }

    /*
     * Calculate the mod value that most closely gives the requested
     * frequency resolution.
     */
    mod = MAX(1, (uint32_t)nearbyint(fref2 / woptp->wopt_fres / (double)r));
    /*FALLTHROUGH*/

success:
    /*
     * For low spur mode, scale mod by an integer such that it's at least
     * ADF435X_MOD_MIN_LS.  Scaling by an integer avoids removing any of
     * the existing factors.
     */
    if (mod < mod_min) {
	mod *= (uint32_t)ceil(mod_min / mod);
    }
    /*
     * The minimum mod value that can be programmed is 2.  If mod is 1,
     * bump it up to 2.
     */
    if (mod < ADF435X_MOD_MIN) {
	mod = ADF435X_MOD_MIN;
    }
    /*
     * Determine the prescalar value, preferring 8/9 when possible.
     */
    temp = floor(vco_fmin / (fref2 / (double)r));
    if (temp < (double)ADF434X_INT_MIN_PRE89) {
	assert(pre45_possible);
	assert(temp >= (double)ADF434X_INT_MIN_PRE45);
	wfsynp->wfsyn_pre8_9 = 0;
    } else {
	wfsynp->wfsyn_pre8_9 = 1;
    }
    /*
     * Now that we know we're going to succeed, set results and commit
     * the new options.
     */
    assert(r >= 1 && r <= ADF435X_R_MAX);
    assert(mod >= ADF435X_MOD_MIN && mod <= ADF435X_MOD_MAX);
    wfsynp->wfsyn_rmul2 = rmul2_enable;
    wfsynp->wfsyn_rdiv2 = rdiv2_enable;
    wfsynp->wfsyn_r = r;
    wfsynp->wfsyn_mod = mod;
    wfsynp->wfsyn_pfd = fref2 / (double)r;
    wfsynp->wfsyn_current_options = wfsynp->wfsyn_new_options;

    /*
     * If the frequency is out-of-range, put it in-range.
     */
    if (wfsynp->wfsyn_frequency < woptp->wopt_fmin ||
	wfsynp->wfsyn_frequency > woptp->wopt_fmax) {
	wfsynp->wfsyn_frequency = woptp->wopt_fmin;
    }

    /*
     * Update the device to the new values.
     */
    return wfsyn_update(wfsynp);

fail:
    wfsyn_error(wfsynp, "wfsyn_set_options: %s: "
			"unachievable parameter combination",
		wfsynp->wfsyn_device_name);
    errno = EINVAL;
    return -1;
}

/*
 * wfsyn_get_pfd: get the computed PFD frequency
 *   @wfsynp: the WFSYN device structure
 */
double wfsyn_get_pfd(const wfsyn_device_type *wfsynp)
{
    return wfsynp->wfsyn_pfd;
}

/*
 * wfsyn_get_rmod: return divider, r, and mod values
 *   @wfsynp: the WFSYN device structure
 *   @rmul2p: optional pointer to returned MUL2 flag
 *   @rdiv2p: optional pointer to returned DIV2 flag
 *   @rp: optional pointer to returned r value
 *   @modp: optional pointer to returned mod value
 */
void wfsyn_get_rmod(const wfsyn_device_type *wfsynp, wfsyn_bool_type *rmul2p,
	   wfsyn_bool_type *rdiv2p, uint16_t *rp, uint16_t *modp)
{
    if (rmul2p != NULL)
	*rmul2p = wfsynp->wfsyn_rmul2;
    if (rdiv2p != NULL)
	*rdiv2p = wfsynp->wfsyn_rdiv2;
    if (rp != NULL)
	*rp = wfsynp->wfsyn_r;
    if (modp != NULL)
	*modp = wfsynp->wfsyn_mod;
}
