/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "wfsyn_internal.h"
#include "wfsyn.h"

/*
 * Make sure consecutive x-terms differ by at least this amount to
 * avoid an ill conditioned matrix.
 */
#define MIN_DX	0.0001

/*
 * wfsyn_spline_calc: find natural cubic spline coefficients
 *   @n:   number of spline segments
 *   @dp:  n+1 element vector of x,y pairs
 *   @scp: n element vector of b,c,d tuples
 *
 * The resulting coefficients go into the following interpolation
 * polynomical:
 *   y(x) = dp[i][Y] + scp[i][B] dx + scp[i][C] dx^2 + scp[i][D] dx^3
 *   where:
 *     dx = x - dp[i][X],
 *     dp[i][X] <= x <= dp[i+1][X]
 */
int wfsyn_spline_calc(int n, const double (*dp)[2], double (*scp)[3])
{
	int i;
	int rv = 0;
	double *hp = NULL;	/* widths of each segment */
	double *mp = NULL;	/* slopes of each segment */
	double *up = NULL;	/* major diagonal of matrix */
	double *vp = NULL;	/* right-hand terms of matrix */
	double *sp = NULL;	/* second derivative at x[i] */

	/*
	 * Allocate temporary vectors.
	 */
	if ((hp = (double *)malloc(n * sizeof(double))) == NULL) {
		rv = -1;
		goto out;
	}
	if ((mp = (double *)malloc(n * sizeof(double))) == NULL) {
		rv = -1;
		goto out;
	}
	if ((up = (double *)malloc(n * sizeof(double))) == NULL) {
		rv = -1;
		goto out;
	}
	if ((vp = (double *)malloc(n * sizeof(double))) == NULL) {
		rv = -1;
		goto out;
	}
	if ((sp = (double *)malloc((n + 1) * sizeof(double))) == NULL) {
		rv = -1;
		goto out;
	}

	/*
	 * Find the segment widths and slopes, and test that the x[i]
	 * terms are increasing.
	 */
	for (i = 0; i < n; ++i) {
		hp[i] =  dp[i+1][X] - dp[i][X];
		mp[i] = (dp[i+1][Y] - dp[i][Y]) / hp[i];
		if (hp[i] < MIN_DX) {
			/* error reported by caller */
			errno = EINVAL;
			return -1;
		}
	}

	/*
	 * Do gaussian elimination, taking advantage of the fact that
	 * the matrix is sparse (tri-diagonal).
	 */
	up[0] = 2.0 * (hp[0] + hp[1]);
	vp[0] = 6.0 * (mp[1] - mp[0]);
	for (i = 1; i < n - 1; ++i) {
		up[i] = 2.0 * (hp[i] + hp[i+1]) - hp[i] * hp[i]   / up[i-1];
		vp[i] = 6.0 * (mp[i+1] - mp[i]) - hp[i] * vp[i-1] / up[i-1];
	}

	/*
	 * Do back-substitution to solve for the second derivative
	 * terms.  The natural cubic spline is defined to have zero
	 * second derivatives at the two endpoints.
	 */
	sp[n] = 0.0;
	for (i = n-1; i >= 1; --i) {
		sp[i] = (vp[i-1] - hp[i] * sp[i+1]) / up[i-1];
	}
	sp[0] = 0.0;

	/*
	 * Compute the coefficients.
	 */
	for (i = 0; i < n; ++i) {
		scp[i][B] = (dp[i+1][Y] - dp[i][Y]) / hp[i]
			      - hp[i] / 3.0 * sp[i]
			      - hp[i] / 6.0 * sp[i+1];
		scp[i][C] = sp[i] / 2.0;
		scp[i][D] = (sp[i+1] - sp[i]) / (6.0 * hp[i]);
	}
	rv = 0;

out:
	/*
	 * Free the temporary vectors.
	 */
	if (sp != NULL)
		free((void *)sp);
	if (vp != NULL)
		free((void *)vp);
	if (up != NULL)
		free((void *)up);
	if (mp != NULL)
		free((void *)mp);
	if (hp != NULL)
		free((void *)hp);

	return rv;
}

/*
 * wfsyn_spline_eval: evaluate the spline at x
 *   @n:   number of spline segments
 *   @dp:  n+1 element vector of x,y pairs
 *   @scp: n element vector of b,c,d tuples
 *   @x:   independent parameter
 */
extern double wfsyn_spline_eval(int n, const double (*dp)[2],
			        const double (*scp)[3], double x)
{
	int i;
	int low, high;
	double dx;

	/*
	 * Special case linear extrapolation on the left.
	 */
	if (x < dp[0][X]) {
		double m; /* first derivative at x[0] */

		m  = scp[0][B];
		return m * (x - dp[0][X]) + dp[0][Y];
	}
	/*
	 * Special case linear extrapolation on the right.
	 */
	if (x >= dp[n][X]) {
		double m; /* first derivative at x[n] */

		dx = dp[n][X] - dp[n-1][X];
		m  = scp[n-1][B] + dx * (2.0 * scp[n-1][C] +
					 dx * 3.0 * scp[n-1][D]);
		return m * (x - dp[n][X]) + dp[n][Y];
	}
	/*
	 * Binary search for the interval containing x and evaluate.
	 */
	low  = 0;
	high = n-1;
	for (;;) {
		i = (low + high) / 2;
		if (low >= high) {
			break;
		}
		if (x <  dp[i][X]) {
			high = i - 1;
			continue;
		}
		if (x >= dp[i+1][X]) {
			low = i + 1;
			continue;
		}
		assert(x >= dp[i][X]);
		assert(x <= dp[i+1][X]);
		break;
	}
	dx = x - dp[i][X];
	return dp[i][Y] + dx * (scp[i][B] +
				dx * (scp[i][C] +
				      dx * scp[i][D]));
}
