/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "wfsyn.h"
#include "wfsyn_internal.h"


/*
 * wfsyn_sweep: perform a sweep from fmin to fmax
 *   @wfsynp: the WFSYN device structure
 *   @fmin: lower frequency in MHz
 *   @fmax: upper frequency in MHz
 *   @nsteps: total number of steps from fmin to fmax
 *   @step_type: selects between linearly or logarithmicly spaced steps
 *   @hold_time: hold time in us
 *   @calibration_vector: optional vector of values to subtract from the result
 *   @frequency_vector: optional vector of resulting frequencies
 *   @result_vector: optional vector of resulting values
 */
int
wfsyn_sweep(wfsyn_device_type *wfsynp, double fmin, double fmax, int nsteps,
	    int step_type, double hold_time, const double *calibration_vector,
	    double *frequency_vector, double *result_vector)
{
    uint64_t ns;
    struct timespec tv;
    double step;
    wfsyn_options_type *woptp = wfsyn_get_options(wfsynp);
    double old_fmin = woptp->wopt_fmin;
    double old_fmax = woptp->wopt_fmax;
    double old_fres = woptp->wopt_fres;
    int rv = 0;

    /*
     * Validate the frequency range.
     */
    if (fmin > fmax) {
	wfsyn_error(wfsynp, "wfsyn_sweep: %s: fmin (%f) > fmax (%f)",
		    wfsynp->wfsyn_device_name, fmin, fmax);

	errno = EINVAL;
	return -1;
    }

    /*
     * Calculate the frequency or log(frequency) step size such
     * that both endpoints are included (if nsteps >= 2).
     */
    if (nsteps < 2) {
	step = 0.0;
    } else if (step_type == WFSYN_STEP_LIN) {
	step = (fmax - fmin)	/ (double)(nsteps - 1);
    } else if (step_type == WFSYN_STEP_LOG) {
	step = log(fmax / fmin) / (double)(nsteps - 1);
    } else {
	wfsyn_error(wfsynp, "wfsyn_sweep: %s: %u: invalid step type",
		    wfsynp->wfsyn_device_name, step_type);
	errno = EINVAL;
	return -1;
    }

    /*
     * Calculate the hold time in nanoseconds and as a struct timespec.
     */
    if (hold_time < 0.0 || hold_time > 2.0e+15) {
	wfsyn_error(wfsynp, "wfsyn_sweep: %s: %f: invalid hold time",
		    wfsynp->wfsyn_device_name, hold_time);
	errno = EINVAL;
	return -1;
    }
    ns = (uint64_t)(hold_time * 1.0e+3);
    tv.tv_sec  = (time_t)(ns / 1000000000L);
    tv.tv_nsec =   (long)(ns % 1000000000L);

    /*
     * Temporarily narrow the frequency range to the sweep range,
     * and the frequency resolution to no more than the smallest
     * step size.
     */
    woptp->wopt_fmin = fmin;
    woptp->wopt_fmax = fmax;
    switch (step_type) {
    case WFSYN_STEP_LIN:
	if (step < woptp->wopt_fres) {
	    woptp->wopt_fres = step;
	}
	break;

    case WFSYN_STEP_LOG:
	if (fmin * (exp(step) - 1.0) < woptp->wopt_fres) {
	    woptp->wopt_fres = fmin * (exp(step) - 1.0);
	}
	break;
    }
    if (wfsyn_set_options(wfsynp) == -1) {
	return -1;
    }

    /*
     * Sweep.
     */
    if (nsteps > 0) {
	int i;

	for (i = 0; i < nsteps; ++i) {
	    double f;

	    /*
	     * Calculate the frequency.
	     */
	    switch (step_type) {
	    case WFSYN_STEP_LIN:
	    default:
		f = fmin + (double)i * step;
		break;

	    case WFSYN_STEP_LOG:
		f = fmin * exp((double)i * step);
		break;
	    }

	    /*
	     * Set the frequency.
	     */
	    if ((rv = wfsyn_set_frequency(wfsynp, f)) == -1)
		goto out;

	    /*
	     * If a hold time was given, wait.
	     */
	    if (ns != 0) {
		(void)nanosleep(&tv, NULL);
	    }

	    /*
	     * If a frequency vector was given, return the frequency.
	     */
	    if (frequency_vector != NULL)
		frequency_vector[i] = wfsyn_get_frequency(wfsynp);

	    /*
	     * If an output vector was given, measure the power detector
	     * value.
	     */
	    if (result_vector != NULL) {
		result_vector[i] = wfsyn_detect_power(wfsynp);
		if (result_vector[i] == HUGE_VAL) {
		    rv = -1;
		    goto out;
		}

		/*
		 * If a calibration vector was given,
		 * subtract the calibration value.
		 */
		if (calibration_vector != NULL)
		    result_vector[i] -= calibration_vector[i];
	    }
	}
    }
    rv = 0;

out:
    /*
     * Restore original fmin, fmax and fres values.
     */
    woptp->wopt_fmin = old_fmin;
    woptp->wopt_fmax = old_fmax;
    woptp->wopt_fres = old_fres;
    (void)wfsyn_set_options(wfsynp);
    return rv;
}
