/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "wfsyn.h"
#include "wfsyn_internal.h"

/*
 * ADF435x output power table
 */
const double adf435x_power_table[] = {
	-4.0, -1.0, +2.0, +5.0
};

/*
 * wfsyn_calc_log2rfdiv: calculate the RF divider shift for the given frequency
 */
int
wfsyn_calc_rfdiv(const wfsyn_device_type *wfsynp, double frequency)
{
    int result;

    result = (wfsynp->wfsyn_capabilities & WFSYN_CAP_ADF4351) ? 6 : 4;
    while (frequency >= ADF435X_VCO_MAX / (double)(1 << result)) {
	if (result == 0) {
	    break;
	}
	--result;
    }
    return result;
}

/*
 * wfsyn_calc_intmod: calculate output divider index, INT and MOD
 *		      for the current target frequency
 */
void
wfsyn_calc_intmod(const wfsyn_device_type *wfsynp,
	 int *ip_rfdiv, uint16_t *u16p_int, uint16_t *u16p_frac)
{
    uint32_t u32;

    *ip_rfdiv = wfsyn_calc_rfdiv(wfsynp, wfsynp->wfsyn_frequency);
    u32 = (uint32_t)nearbyint(wfsynp->wfsyn_frequency *
                              (double)(1U << *ip_rfdiv) *
		              (double)wfsynp->wfsyn_mod /
			      wfsynp->wfsyn_pfd);
    assert(u32 <= ADF435X_INT_MAX * ADF435X_MOD_MAX);
    *u16p_int  = u32 / wfsynp->wfsyn_mod;
    *u16p_frac = u32 % wfsynp->wfsyn_mod;
}

/*
 * wfsyn_calc_power: calculate adf435x and hmc625 power values for dBm_power
 *   @power: desired output power in dBm
 *   @u8_adf435x: pointer to RF-enable bit and two output-power bits
 *   @u8_hmc625:  pointer to HBC625 control byte
 */
void
wfsyn_calc_power(const wfsyn_device_type *wfsynp,
		 uint8_t *u8_adf435x, uint8_t *u8_hmc625)
{
    double power = wfsynp->wfsyn_power;
    double temp;

    /*
     * If power == WFSYN_MUTE, disable the output.
     */
    if (WFSYN_POWER_IS_MUTED(power)) {
	*u8_adf435x = 0x0;  /* RF disabled */
	*u8_hmc625  = 0x40;
	return;
    }

    /*
     * Add the calibration offset and bound.
     */
    power += wfsynp->wfsyn_power_offset;
    if (power < adf435x_power_table[0] + HMC625_POWER0)
	power = adf435x_power_table[0] + HMC625_POWER0;
    else if (power > adf435x_power_table[3] + HMC625_POWER63)
	power = adf435x_power_table[3] + HMC625_POWER63;

    /*
     * Low-range: -17.5 to -9.0 dBm
     */
    if (power <= -8.75) {
	temp = nearbyint((power - (adf435x_power_table[0] + HMC625_POWER0)) /
			 HMC625_POWER_STEP);
	assert(temp >= 0.0 && temp <= 63.0);
	*u8_adf435x = 0x4;  /* RF enabled, -4 dB */
	*u8_hmc625 = 0xC0 | (uint8_t)temp;
	return;
    }
    /*
     * High-range: -8.5 to +23.0 dBm
     */
    temp = nearbyint((power - (adf435x_power_table[3] + HMC625_POWER0)) /
	             HMC625_POWER_STEP);
    assert(temp >= 0.0 && temp <= 63.0);
    *u8_adf435x = 0x7;	/* RF enabled, +5 dB */
    *u8_hmc625 = 0xC0 | (uint8_t)temp;
    return;
}

/*
 * wfsyn_update: update the device registers
 *   @wfsynp: the WFSYN device structure
 */
int
wfsyn_update(wfsyn_device_type *wfsynp)
{
    const wfsyn_options_type *woptp = &wfsynp->wfsyn_current_options;
    uint32_t adf435x[6];	/* new ADF435X registers */
    uint8_t rfen_power = 0;	/* ADF435x RF enable and RF power */
    uint8_t hmc625 = 0;		/* new HMC625 value */
    int rfdiv;			/* log2(RF divider) */
    uint16_t INT;		/* ADF435x integer part */
    uint16_t FRAC;		/* ADF435x modulus part */
    double temp;
    uint8_t adf435x_modified = 0;
    char buffer[WFSYN_MAX_LINE + 1];
    int i;

    /*
     * Init new register values to just their control bits.
     */
    (void)memset((void *)adf435x, 0, sizeof(adf435x));
    for (i = 0; i < 6; ++i)
	adf435x[i] = i;

    /*
     * Calculate RF-divider, INT and FRAC values for the selected
     * frequency.
     */
    wfsyn_calc_intmod(wfsynp, &rfdiv, &INT, &FRAC);

    /*
     * Calculate power settings for the current power.	Note that
     * rfen_power contains both the RF enable and output power
     * bits.
     */
    wfsyn_calc_power(wfsynp, &rfen_power, &hmc625);

    /*
     * Insert INT and FRAC into R0.
     */
    adf435x[0] = ((uint32_t)INT	 << ADF435X_R0_INT_SHIFT) |
	         ((uint32_t)FRAC << ADF435X_R0_FRAC_SHIFT);

    /*
     * Set the prescalar value in R1.
     */
    if (wfsynp->wfsyn_pre8_9)
	adf435x[1] |= ADF435X_R1_PRESCALER;

    /*
     * Set phase shift and modulus in R1.
     */
    adf435x[1] |= (1 << ADF435X_R1_PHASE_SHIFT) |
	          (wfsynp->wfsyn_mod << ADF435X_R1_MOD_SHIFT);

    /*
     * Set low noise/low spur mode in R2.
     */
    switch (woptp->wopt_mode) {
    case WFSYN_MODE_LOW_SPUR:
	adf435x[2] |= 0x3 << ADF435X_R2_LNLS_SHIFT;
	break;
    case WFSYN_MODE_FRAC_N:
    case WFSYN_MODE_INT_N:
	break;
    }

    /*
     * Set MUXOUT to reflect digital lock status in R2.
     */
    adf435x[2] |= ADF435X_MUX_NDLD << ADF435X_R2_MUXOUT_SHIFT;

    /*
     * Set RMUL2 and RDDIV2 flags in R2.
     */
    if (wfsynp->wfsyn_rmul2)
	adf435x[2] |= ADF435X_R2_RMUL2;
    if (wfsynp->wfsyn_rdiv2)
	adf435x[2] |= ADF435X_R2_RDIV2;

    /*
     * Set R, double-buffer mode and charge pump current in R2.
     */
    adf435x[2] |= (wfsynp->wfsyn_r << ADF435X_R2_R_SHIFT) |
		  ADF435X_R2_DOUBLE_BUFFER |
		  (woptp->wopt_icp << ADF435X_R2_ICP_SHIFT);
    switch (woptp->wopt_mode) {
    case WFSYN_MODE_INT_N:
	adf435x[2] |= ADF435X_R2_LDF;
	break;
    case WFSYN_MODE_FRAC_N:
    case WFSYN_MODE_LOW_SPUR:
	break;
    }

    /*
     * Set the phase detector polarity to positive.
     */
    adf435x[2] |= ADF435X_R2_PD_POLARITY;

    /* TODO: implement power down or VCO power down? */

    /*
     * Set R2 options specific to ADF4351.
     */
    if (wfsynp->wfsyn_capabilities & WFSYN_CAP_ADF4351) {
	/*
	 * Enable fast band select clock mode provided that PFD is
	 * greater than 125 kHz.
	 */
	if (wfsynp->wfsyn_pfd > 0.125)
		adf435x[3] |= ADF435X_R3_BSCM;

	/*
	 * In INT-N mode, set antibacklash pulse width to 3ns and
	 * charge cancel mode.
	 */
	switch (woptp->wopt_mode) {
	case WFSYN_MODE_INT_N:
	    adf435x[3] |= ADF435X_R3_ABP | ADF435X_R3_CHARGE_CANCEL;
	    break;

	case WFSYN_MODE_FRAC_N:
	case WFSYN_MODE_LOW_SPUR:
	    break;
	}
    }

    /*
     * Set the 12-bit clock divider value for phase resync and fast lock
     * modes.  Currently, this is hard-coded because we're not using those
     * modes, but if this changes, it should be calculated as:
     *
     *	nearbyint(PFD / MOD * (20us + delay)).
     */
    adf435x[3] |= (150 << ADF435X_R3_CLKDIV_VALUE_SHIFT);

    /*
     * Set fundamental feedback mode in R4.
     */
    adf435x[4] |= ADF435X_R4_FEEDBACK_SELECT;

    /*
     * Set the RF divider in R4.
     */
    adf435x[4] |= rfdiv << ADF435X_R4_RF_DIVIDER_SHIFT;

    /*
     * Program the band select clock divider for a band select
     * frequency less than or equal to 125 kHz.
     */
    temp = ceil(woptp->wopt_fref * (wfsynp->wfsyn_rmul2 ? 2.0 : 1.0) /
	    wfsynp->wfsyn_r / ADF435X_BSFMAX);
    adf435x[4] |= (MIN(0xff, (uint32_t)temp) << ADF435X_R4_BSCDV_SHIFT);

    /*
     * Set mute until lock detect in R4.
     */
    adf435x[4] |= ADF435X_R4_MTLD;

    /*
     * Set RF enable and RF power in R4.
     */
    adf435x[4] |= rfen_power << ADF435X_R4_RFOUT_POWER_SHIFT;

    /*
     * Set the lock detect pin to digital in R5.
     */
    adf435x[5] |= 1 << ADF435X_R5_LDPIN_MODE_SHIFT;

    /*
     * Set the reserved to 1 bits in R5.
     */
    adf435x[5] |= ADF435X_R5_RESERVED_ONES;

    /*
     * Create a bitmask of which ADF435X registers actually changed.
     */
    for (i = 0; i < 6; ++i) {
	if (adf435x[i] != wfsynp->wfsyn_adf435x[i]) {
		adf435x_modified |= (1U << i);
	}
    }

    /*
     * For double buffering, a change to R1, R2 or R4 implies that
     * R0 must also be written.
     */
    if (adf435x_modified & 0x16) {
	adf435x_modified |= 0x1;
    }

    /*
     * Do the update for firmware that suppports text commands.
     */
    if (wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD) {
	/*
	 * If the HMC625 amplitude became smaller, update it
	 * before the ADF435x registers.
	 */
	if (hmc625 < wfsynp->wfsyn_hmc625) {
	    if (wfsyn_printf(wfsynp, "a%u", hmc625) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_hmc625 = hmc625;
	}

	/*
	 * Update all modified HMC625 registers, writing them in reverse
	 * order.  By sending all 8 hex digits, we don't have to worry
	 * about ambiguity if the next command in the byte stream begins
	 * with [a-fA-F].
	 */
	for (i = 5; i >= 0; --i) {
	    if (adf435x_modified & (1U << i)) {
		if (wfsyn_printf(wfsynp, "H%u%08X",
			         i, adf435x[i]) == -1) {
		    return -1;
		}
		wfsynp->wfsyn_adf435x[i] = adf435x[i];
	    }
	}

	/*
	 * If the HMC625 amplitude became larger, update it
	 * after the ADF435x registers.
	 */
	if (hmc625 > wfsynp->wfsyn_hmc625) {
	    if (wfsyn_printf(wfsynp, "a%u", hmc625) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_hmc625 = hmc625;
	}

	/*
	 * If the internal reference flag changed, update it.
	 */
	if (woptp->wopt_refsrc != wfsynp->wfsyn_refsrc) {
	    if (wfsyn_printf(wfsynp, "x%u", woptp->wopt_refsrc) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_refsrc = woptp->wopt_refsrc;
	}

    /*
     * Update using the binary interface.
     */
    } else {
	/*
	 * If the HMC625 amplitude became smaller, update it
	 * before the ADF435x registers.
	 */
	if (hmc625 < wfsynp->wfsyn_hmc625) {
	    buffer[0] = 0x03;
	    buffer[1] = hmc625;
	    if (wfsyn_write(wfsynp, buffer, 2) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_hmc625 = hmc625;
	}

	/*
	 * If only R0 changed, send a single register update.
	 */
	if (adf435x_modified == 0x01) {
	    buffer[0] = 0x02;
	    buffer[1] = (adf435x[0] >> 24) & 0xff;
	    buffer[2] = (adf435x[0] >> 16) & 0xff;
	    buffer[3] = (adf435x[0] >>  8) & 0xff;
	    buffer[4] =  adf435x[0]        & 0xff;
	    if (wfsyn_write(wfsynp, buffer, 5) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_adf435x[0] = adf435x[0];

	/*
	 * If R1..R5 changed, write all six registers.
	 */
	} else if (adf435x_modified != 0x00) {
	    int j = 0;

	    buffer[j++] = 0x01;
	    for (i = 0; i < 6; ++i) {
		buffer[j++] = (adf435x[i] >> 24) & 0xff;
		buffer[j++] = (adf435x[i] >> 16) & 0xff;
		buffer[j++] = (adf435x[i] >>  8) & 0xff;
		buffer[j++] =  adf435x[i]        & 0xff;
	    }
	    if (wfsyn_write(wfsynp, buffer, 25) == -1) {
		return -1;
	    }
	    /*
	     * The V1 firmware writes the registers in the order R0, R1, ...
	     * R5.  According to the ADF435x data sheet, they should be written
	     * in R5, R4, ... R0 order.  So the double-buffered settings don't
	     * go into effect.  Write R0 again as a work-around.
	     */
	    buffer[0] = 0x02;
	    buffer[1] = (adf435x[0] >> 24) & 0xff;
	    buffer[2] = (adf435x[0] >> 16) & 0xff;
	    buffer[3] = (adf435x[0] >>  8) & 0xff;
	    buffer[4] =  adf435x[0]        & 0xff;
	    if (wfsyn_write(wfsynp, buffer, 5) == -1) {
		return -1;
	    }
	    for (i = 0; i < 6; ++i)
		wfsynp->wfsyn_adf435x[i] = adf435x[i];
	}

	/*
	 * If the HMC625 amplitude became larger, update it
	 * after the ADF435x registers.
	 */
	if (hmc625 > wfsynp->wfsyn_hmc625) {
	    buffer[0] = 0x03;
	    buffer[1] = hmc625;
	    if (wfsyn_write(wfsynp, buffer, 2) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_hmc625 = hmc625;
	}

	/*
	 * If the internal reference flag changed, update it.
	 */
	if (woptp->wopt_refsrc != wfsynp->wfsyn_refsrc) {
	    buffer[0] = 0x06;
	    buffer[1] = woptp->wopt_refsrc;
	    if (wfsyn_write(wfsynp, buffer, 2) == -1) {
		return -1;
	    }
	    wfsynp->wfsyn_refsrc = woptp->wopt_refsrc;
	}
    }

    /*
     * If the ADF435x registers were modified, wait for lock.  Without
     * this, we see glitches when sweeping or when changing the power
     * between the low and high ranges.
     */
    if (adf435x_modified != 0) {
	struct timespec tv;

	/*
	 * If the wlock option is set, actively poll-wait with Fibonacci
	 * backoff.
	 */
	if (woptp->wopt_wlock) {
	    uint32_t usec0 = 100;	/* first wait is 100us */
	    uint32_t usec1 = 0;
	    uint32_t temp;
	    uint32_t total = 0;

	    assert(wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD);
	    for (;;) {
		char buffer[5];

		if (wfsyn_printf(wfsynp, "p") == -1) {
		    return -1;
		}
		if (wfsyn_readline(wfsynp, buffer, sizeof(buffer)) == -1) {
		    return -1;
		}
		if (atoi(buffer) != 0) {
		    break;
		}
		if (total >= 1000000) {	/* time-out after 1s */
		    wfsyn_error(wfsynp, "wfsyn_set_frequency: %s: "
					"PLL failed to lock",
				wfsynp->wfsyn_device_name);
		    errno = ERANGE;
		    return -1;
		}
		tv.tv_sec  =  usec0 / 1000000;
		tv.tv_nsec = (usec0 % 1000000) * 1000;
		(void)nanosleep(&tv, NULL);
		total += usec0;
		temp   = usec0;
		usec0 += usec1;
		usec1  = temp;
	    }

	/*
	 * Otherwise wait for a fixed delay.
	 */
	} else {
	    tv.tv_sec  = 0;
	    tv.tv_nsec = 7000000;	/* 7 ms */
	    (void)nanosleep(&tv, NULL);
	}
    }
    return 0;
}

/*
 * wfsyn_save_to_nvram: save settings to NVRAM
 */
int
wfsyn_save_to_nvram(wfsyn_device_type *wfsynp)
{
    char buffer[1];

    /*
     * In the new interface, first set the frequency.  The "f" command
     * clobbers some of the device registers.  Then, force an update of
     * all registers by clearing the shadowed versions to zeros (which
     * is an invalid value for all registers) and calling update.  With the
     * registers restored, we can send the "e" command to save everything.
     */
    if (wfsynp->wfsyn_capabilities & WFSYN_CAP_TEXTCMD) {
	if (wfsyn_printf(wfsynp, "f%6.4f", wfsynp->wfsyn_frequency) == -1) {
	    return -1;
	}
	(void)memset((void *)wfsynp->wfsyn_adf435x, 0,
		     sizeof(wfsynp->wfsyn_adf435x));
	if (wfsyn_update(wfsynp) == -1) {
	    return -1;
	}
	return wfsyn_printf(wfsynp, "e");
    }

    /*
     * In the old interface, all we have to do is send command 5.
     */
    buffer[0] = '\005';
    return wfsyn_write(wfsynp, buffer, 1);
}
