/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "wfsyn_internal.h"
#include "wfsyn.h"
#include "command.h"
#include "calibrate_detector.h"
#include "main.h"

#define MAX_LINE	80

/*
 * cmd_write: write a string to a file descriptor
 */
static int
cmd_write(int fd, char *buffer, int size)
{
    char *cur = buffer;
    int rv;

    while (size > 0) {
	if ((rv = write(fd, cur, size)) == -1) {
	    return -1;
	}
	cur  += rv;
	size -= rv;
    }
    return cur - buffer;
}


/*
 * cmd_readline: read a line from a file descriptor
 */
static int
cmd_readline(int fd, char *buffer, int size)
{
    char *cur = buffer;
    int rv;

    while (size > 0) {
	if ((rv = read(fd, cur, size)) == -1) {
	    return -1;
	}
	if (rv == 0) {
	    break;
	}
	cur  += rv;
	size -= rv;
	if (cur[-1] == '\n') {
	    break;
	}
    }
    return cur - buffer;
}


/*
 * parse_response: parse an interactive response string
 *   @powerp: modiifed if a value is given
 *
 * Return:
 *   RESP_OK:    ok or new value
 *   RESP_SKIP:  skip
 *   RESP_QUIT:  quit
 *   RESP_ERROR: invalid response
 */
static response_type
parse_response(char *response, double *powerp)
{
    char *cp;
    double lf;
    const char whitespace[] = " \t\f\v\r";

    /*
     * Skip leading whitespace (excluding newline).
     */
    response += strspn(response, whitespace);

    /*
     * Remove trailing whitespace.
     */
    cp = response + strlen(response);
    while (cp > response && strchr(whitespace, cp[-1]) != NULL)
	*--cp = '\000';

    /*
     * Handle each response.
     */
    switch (response[0]) {
    case '\000':
	return RESP_OK;

    case 'o':
	if (strncmp(response, "ok", strlen(response)) == 0) {
	    return RESP_OK;
	}
	break;

    case 's':
	if (strncmp(response, "skip", strlen(response)) == 0) {
	    return RESP_SKIP;
	}
	break;

    case 'q':
	if (strncmp(response, "quit", strlen(response)) == 0) {
	    return RESP_QUIT;
	}
	break;

    default:
	/*
	 * If a valid floating number is given, override the
	 * offered value with the response.
	 */
	lf = strtod(response, &cp);
	if (*cp == '\000') {
	    *powerp = lf;
	    return RESP_OK;
	}
	break;
    }
    (void)fprintf(stderr, "%s: invalid input\n", progname);
    return RESP_ERROR;
}

/*
 * set_power_c: set the power using a non-interactive script
 */
response_type
set_power_c(double *powerp, const char *cmd)
{
    char buffer[MAX_LINE + 1];
    char *cp;
    int recv_fd;
    pid_t pid;
    int rv;

    (void)sprintf(buffer, "%8.3f", *powerp);
    if (cmd_start1(cmd, buffer, &recv_fd, &pid) == -1) {
	return RESP_ERROR;
    }
    rv = cmd_readline(recv_fd, buffer, sizeof(buffer));
    if (rv == -1) {
	(void)fprintf(stderr, "%s: read: %s\n", progname, strerror(errno));
	(void)cmd_wait(pid);
	return RESP_ERROR;
    }
    (void)close(recv_fd);
    if (cmd_wait(pid) == -1) {
	return RESP_ERROR;
    }
    if (rv == 0) {	/* EOF OK */
	return RESP_OK;
    }
    cp = buffer + rv;
    if (cp[-1] != '\n') {
	(void)fprintf(stderr, "%s: invalid response\n", progname);
	return RESP_ERROR;
    }
    *--cp = '\000';
    return parse_response(buffer, powerp);
}

/*
 * set_power_ci: set the power using an interactive script
 */
response_type
set_power_ci(double *powerp, int *fdp)
{
    char buffer[MAX_LINE + 1];
    char *cp;
    int rv;

    (void)sprintf(buffer, "%8.3f\n", *powerp);
    if ((rv = cmd_write(fdp[1], buffer, strlen(buffer))) == -1) {
	(void)fprintf(stderr, "%s: write: %s\n", progname, strerror(errno));
	return RESP_ERROR;
    }
    if ((rv = cmd_readline(fdp[0], buffer, sizeof(buffer))) == -1) {
	(void)fprintf(stderr, "%s: read: %s\n", progname, strerror(errno));
	return RESP_ERROR;
    }
    if (rv == 0) {
	(void)fprintf(stderr, "%s: unexpected EOF from command\n", progname);
	return RESP_ERROR;
    }
    cp = buffer + rv;
    if (cp[-1] != '\n') {
	(void)fprintf(stderr, "%s: invalid response\n", progname);
	return RESP_ERROR;
    }
    *--cp = '\000';
    return parse_response(buffer, powerp);
}

/*
 * set_power_i: prompt the user to set the reference power manually
 *   @powerp: requested power (in); actual power (out)
 */
response_type
set_power_i(double *powerp)
{
    char buffer[MAX_LINE + 1];
    response_type r;
    char *cp;
    int i;

    for (;;) {
	(void)printf("%8.3f dBm {ok|skip|quit|<value>} ", *powerp);
	if (fgets(buffer, sizeof(buffer), stdin) == NULL) {
	    return RESP_QUIT;
	}
	if ((cp = strchr(buffer, '\n')) == NULL) {
	    while ((i = getchar()) != '\n') {
		if (i == EOF)
		    break;
	    }
	    (void)fprintf(stderr, "%s: invalid response\n", progname);
	    continue;
	}
	*cp = '\000';
	if ((r = parse_response(buffer, powerp)) != RESP_ERROR) {
	    return r;
	}
    }
}

/*
 * Number of times to read the power detector.
 */
#define N_TRIALS	35

/*
 * measure_power: measure the power and calculate sum of squares from mean
 *   @wfsynP:  Synth NV device handle
 *   @y:       reference power value
 *   @mp:      resulting measurement struct
 *   @s2p:     cumulative sum of squares from measurement means
 */
void measure_power(wfsyn_device_type *wfsynp, double y,
		   measurement_type *mp, double *s2p)
{
    int i;
    double sx  = 0.0;		/* sum of x's */
    double sx2 = 0.0;		/* sum of x^2's */

    /*
     * Wait 10ms for the detector to stabilize.
     */
    usleep(10000);

    /*
     * Collect ntrials measurements and calculate the mean of the x's.
     */
    for (i = 0; i < N_TRIALS; ++i) {
	double x = wfsyn_detect_power_raw(wfsynp);

	/*
	 * If verbose 3, print the raw measurements.
	 */
	if (opt_v >= 3) {
	    (void)printf("%3.0f %8.3f\n", x, y);
	}
	sx  += x;
	sx2 += x * x;
    }
    mp->m_x = sx / (double)N_TRIALS;		/* mean */
    mp->m_y = y;
    *s2p += sx2 - sx * sx / (double)N_TRIALS;	/* variance */
}

/*
 * m_cmp: qsort compare function to sort measurement_type by m_y
 */
static int m_cmp(const void *vp1, const void *vp2)
{
    const measurement_type *mp1 = vp1;
    const measurement_type *mp2 = vp2;

    if (mp1->m_y < mp2->m_y)
	return -1;
    if (mp1->m_y > mp2->m_y)
	return +1;
    return 0;
}


/*
 * test_spline: test if any point along a spline is non-increasing
 *   @n:   number of spline segments
 *   @dp:  n+1  x,y data pairs
 *   @scp: n-element vector of b,c,d tuples
 *
 * The spline interpolation polynomial is:
 *   y(x) = dp[i][Y] + scp[i][B] dx + scp[i][C] dx^2 + scp[i][D] dx^3
 *   where:
 *     dx = x - dp[i][X]
 *     dp[i][X] <= x <= dp[i+1][X]
 */
static int
test_spline(int n, const double (*dp)[2], const double (*scp)[3])
{
    int i;

    for (i = 0; i < n; ++i) {

	/*
	 * If the B coefficient (first derivative at dp[i][X]) is non-positive,
	 * then the spline is non-increasing at the beginning of the interval.
	 */
	if (scp[i][B] <= 0.0) {
	    return 1;
	}

	/*
	 * Else if the D coefficient is non-zero, then the derivative of
	 * the spline has two roots:
	 *
	 *		-c +/- sqrt(c^2 - 3 * d * b)
	 *		----------------------------
	 *		           3 * d
	 */
	if (scp[i][D] < -EPS || scp[i][D] > EPS) {
	    double temp = scp[i][C] * scp[i][C] - 3.0 * scp[i][D] * scp[i][B];
	    double x1, x2;

	    /*
	     * If the roots are complex, skip this interval.
	     */
	    if (temp < 0.0) {
		continue;
	    }

	    /*
	     * Return true if there is a root between dp[i][X] and dp[i+1][X].
	     */
	    temp = sqrt(temp);
	    x1 = (-scp[i][C] - temp) / (3.0 * scp[i][D]);
	    x2 = (-scp[i][C] + temp) / (3.0 * scp[i][D]);
	    if ((dp[i][X] <= x1 && x1 <= dp[i+1][X]) ||
		(dp[i][X] <= x2 && x2 <= dp[i+1][X])) {
		return 1;
	    }
	    continue;
	}

	/*
	 * Else if the C coefficient is non-zero, then the derivative has
	 * one root: x = -b / (2 * c)
	 */
	if (scp[i][C] < -EPS || scp[i][C] > EPS) {
	    double x = -scp[i][B] / (2.0 * scp[i][C]);

	    if (dp[i][X] <= x && x <= dp[i+1][X]) {
		return 1;
	    }
	    continue;
	}
    }
    return 0;
}

/*
 * calibrate_detector: calibrate the power detector based on measurements
 *   @wfsynP: Synth NV device handle
 *   @mp: vector of ADC-reference measurements
 *   @n_measurements: number of measurements in mp
 */
void calibrate_detector(wfsyn_device_type *wfsynp, measurement_type *mp,
			int n_measurements, double width)
{
    double (*dp)[2];
    double (*scp)[3];
    double next_width = 0.0;
    int i;
    int n_points;

    assert(n_measurements > 1);

    /*
     * If verbose 2, print the averaged measurements.
     */
    if (opt_v == 2) {
	(void)printf("# averaged power measurements\n");
	for (i = 0; i < n_measurements; ++i) {
	    (void)printf("m %7.3f %8.3f\n",
		mp[i].m_x, mp[i].m_y);
	}
	(void)printf("\n\n");
    }

    /*
     * Calculate the initial number of bins from dx and allocate
     * data and coefficient vectors for the spline.
     */
    if ((dp = malloc(n_measurements * sizeof(double[2]))) == NULL) {
	(void)fprintf(stderr, "%s: malloc: %s\n", progname, strerror(errno));
	exit(3);
	/*NOTREACHED*/
    }
    if ((scp = malloc((n_measurements - 1) * sizeof(double[3]))) == NULL) {
	(void)fprintf(stderr, "%s: malloc: %s\n", progname, strerror(errno));
	exit(4);
	/*NOTREACHED*/
    }

    /*
     * Sort the measurements by increasing reference level.
     */
    qsort(mp, n_measurements, sizeof(measurement_type), m_cmp);

    /*
     * Loop until we find an always-increasing spline (or fail to do so).
     */
    for (;;) {
	int count;
	double x0;

	if (opt_v > 0) {
	    (void)printf("# trying bin width %f\n", width);
	}
	n_points = 0;
	dp[n_points][X] = mp[0].m_x;
	dp[n_points][Y] = mp[0].m_y;
	x0 = dp[n_points][X];
	count = 1;
	for (i = 1; i < n_measurements; ++i) {
	    double dx = mp[i].m_x - x0;

	    if (dx > width) {
		if (i == 1 || dx < next_width) {
		    next_width = dx;
		}
		dp[n_points][X] /= (double)count;
		dp[n_points][Y] /= (double)count;
		++n_points;
		dp[n_points][X] = mp[i].m_x;
		dp[n_points][Y] = mp[i].m_y;
		x0 = dp[n_points][X];
		count = 1;
		continue;
	    }
	    dp[n_points][X] += mp[i].m_x;
	    dp[n_points][Y] += mp[i].m_y;
	    ++count;
	}
	if (count > 0) {
	    dp[n_points][X] /= (double)count;
	    dp[n_points][Y] /= (double)count;
	    ++n_points;
	    count = 0;
	}

	if (n_points < 2) {
	    (void)fprintf(stderr, "%s: invalid calibration data\n", progname);
	    exit(3);
	    /*NOTREACHED*/
	}

	/*
	 * Calculate the spline coefficients and test for
	 * non-increasing derivative.
	 */
	if (wfsyn_spline_calc(n_points - 1,
			      (const double (*)[2])dp, scp) == 0) {
	    if (test_spline(n_points - 1, (const double (*)[2])dp,
			    (const double (*)[3])scp) == 0) {
		break;		/* spline is always increasing */
	    }
	    if (opt_v > 0) {
		(void)printf("# spline contains zero-derivatives\n");
	    }
	} else {
	    if (opt_v > 0) {
		(void)printf("# measurements are not increasing\n");
	    }
	    if (errno != EINVAL) {
		(void)fprintf(stderr, "%s: wfsyn_spline_calc: %s\n", progname,
			      strerror(errno));
		exit(4);
		/*NOTREACHED*/
	    }
	}
	width = MAX(width + MIN_DX, next_width);
    }
    if (opt_v > 0) {
	(void)printf("# final spline points\n");
	for (i = 0; i < n_points; ++i) {
	    (void)printf("%10.6f %11.6f\n", dp[i][X], dp[i][Y]);
	}
	(void)printf("\n\n");
    }
    (void)wfsyn_set_detector_calibration(wfsynp, n_points, dp);

    free((void *)scp);
    scp = NULL;
    free((void *)dp);
    dp = NULL;
}
