/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <errno.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "wfsyn.h"
#include "calibrate_output.h"
#include "main.h"

/*
 * calibrate_output: calibrate the output power
 *   @wfsynP: Synth NV device handle
 *   @attenuation: value of attenuator between output and detector
 *
 * This function calibrates the Windfreak Synth NV output power using
 * the Synth NV power detector, which should be calibrated first.
 */
void
calibrate_output(wfsyn_device_type *wfsynp, double attenuation)
{
    int nmax, i;
    int n = 0;
    double old_offset = wfsyn_get_power_offset(wfsynp);
    double *hann_vector;
    double error;
    double scale;
    double sx  = 0.0;	/* sum of x's */
    double sy  = 0.0;	/* sum of y's */
    double sx2 = 0.0;	/* sum of x^2's */
    double sxy = 0.0;	/* sum of xy's */
    double d;		/* denominator */
    double m, b;	/* slope and intercept */


    /*
     * Set the power offset to zero.
     */
    (void)wfsyn_set_power_offset(wfsynp, 0.0);

    /*
     * Calculate the number of output power levels.
     */
    nmax = (int)rint((POWER_MAX - POWER_MIN) / POWER_STEP) + 1;

    /*
     * Create a Hann function used to weight the measurements below.
     */
    if ((hann_vector = malloc(nmax * sizeof(double))) == NULL) {
	(void)fprintf(stderr, "%s: malloc: %s\n", progname, strerror(errno));
	exit(2);
	/*NOTREACHED*/
    }
    for (i = 0; i < nmax; ++i)
	hann_vector[i] = 1.0 - cos(2 * PI * i / (nmax - 1)) / 2.0;

    /*
     * If verbose 2, print a header.
     */
    if (opt_v >= 2) {
	(void)printf("# column1: requested power\n");
	(void)printf("# column2: measured  power\n");
    }

    /*
     * For each output power level, use the (calibrated) detector to measure
     * the power error.
     */
    error = 0.0;
    scale = 0.0;
    for (i = 0; i < nmax; ++i) {
	double requested_power = POWER_MIN + (double)i * POWER_STEP;
	double measured_power;

	/*
	 * Skip power levels that would exceed MAX_DETECTOR_POWER.
	 */
	if (requested_power - attenuation > MAX_DETECTOR_POWER)
	    break;

	/*
	 * Set the power level and measure the error.  Sleep 10ms
	 * to allow the detector to stabilize before each measurement.
	 */
	assert(requested_power - attenuation <= MAX_DETECTOR_POWER);
	(void)wfsyn_set_power(wfsynp, requested_power);
	usleep(10000);
	measured_power = wfsyn_detect_power(wfsynp) + attenuation;
	if (opt_v >= 2) {
	    printf("%7.3f %7.3f\n", requested_power, measured_power);
	}

	/*
	 * Collect stats for linear regression.
	 */
	sx  += requested_power;
	sy  += measured_power;
	sx2 += requested_power * requested_power;
	sxy += requested_power * measured_power;
	++n;

	/*
	 * Accumulate the weighted sum of errors and sum of weights.
	 */
	error += hann_vector[i] * (measured_power - requested_power);
	scale += hann_vector[i];
    }
    error /= scale;	/* normalize */

    /*
     * Compute slope and intercept of the unweighted data.
     */
    d = sx2 - sx * sx / (double)n;
    if (fabs(d) < EPS) {
	(void)fprintf(stderr, "%s: invalid calibration data\n", progname);
	exit(2);
	/*NOTREACHED*/
    }
    m = (sxy - sx * sy / (double)n) / d;
    b = (sy - m * sx) / (double)n;
    if (opt_v > 0) {
	(void)printf("# unweighted regression line: actual = "
	             "%f * requested + %f dBm\n", m, b);
	(void)printf("#\n");
    }

    /*
     * Fail if the slope is not within 20% of 1.
     */
    if (m < 0.80 || m > 1.20) {
	(void)fprintf(stderr, "%s: invalid calibration data\n", progname);
    }

    /*
     * If verbose, print a summary of old and new power offset setting.
     */
    if (opt_v > 0) {
	(void)printf("# old power offset %7.3f\n", old_offset);
	(void)printf("# new power offset %7.3f\n", -error);
    }

    /*
     * Set the new power offset and save.
     */
    (void)wfsyn_set_power_offset(wfsynp, -error);

    free((void *)hann_vector);
}
