/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/wait.h>
#include <assert.h>
#include <errno.h>
#include <getopt.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "command.h"
#include "main.h"


#define MAX_ARGS	4

/*
 * cmd_exec: exec the given command
 *   @cmdp: command (may be simple or a /bin/sh script)
 *   @argp: optional argument to the command (NULL for none)
 */
int
cmd_exec(const char *cmdp, const char *argp)
{
    int argc = 0;
    char *argv[MAX_ARGS + 1];

    /*
     * If the command contains any shell metacharacters, run it under
     * /bin/sh -c.
     */
    if (strpbrk(cmdp, " \t\n!$&'()*:;<>?[\"\\]^`{|}~") != NULL) {
	argv[argc++] = "/bin/sh";
	argv[argc++] = "-c";
    }
    argv[argc++] = (char *)cmdp;
    if (argp != NULL) {
	argv[argc++] = (char *)argp;
    }
    argv[argc] = NULL;
    (void)execvp(argv[0], argv);
    (void)fprintf(stderr, "%s: execvp: %s: %s\n",
    		  progname, argv[0], strerror(errno));
    return -1;
}

/*
 * cmd_wait: wait for the command to exit
 */
int
cmd_wait(pid_t pid)
{
    int status = -1;
    int rv = -1;

    if (waitpid(pid, &status, 0) == -1) {
	(void)fprintf(stderr, "%s: waitpid: %s\n",
		      progname, strerror(errno));
	return -1;
    }
    if (WIFSIGNALED(status)) {
	(void)fprintf(stderr, "%s: command received signal %d\n",
		      progname, WTERMSIG(status));
	return -1;
    }
    if (!WIFEXITED(status) || (rv = WEXITSTATUS(status) != 0)) {
	(void)fprintf(stderr, "%s: command exited %d\n", progname, rv);
	return -1;
    }
    return 0;
}

/*
 * cmd_start1: form a one-way pipe from the given command
 *   @cmdp: command (may be simple or a /bin/sh command)
 *   @argp: argument to command
 *   @fdp:  pointer to write descriptor
 *   @pipd: process ID of child
 */
int
cmd_start1(const char *cmdp, const char *argp, int *fdp, pid_t *pidp)
{
    int recv_pipe[2];

    if (pipe(recv_pipe) == -1) {
	  (void)fprintf(stderr, "%s: pipe: %s\n", progname, strerror(errno));
	  return -1;
    }
    switch (*pidp = fork()) {
    case -1:
	(void)fprintf(stderr, "%s: fork: %s\n", progname, strerror(errno));
	return -1;

    case 0:	/* child */
	if (dup2(recv_pipe[1], 1) == -1) {
	    (void)fprintf(stderr, "%s: dup2: %s\n", progname, strerror(errno));
	    exit(99);
	    /*NOTREACHED*/
	}
	(void)close(recv_pipe[0]);
	(void)close(recv_pipe[1]);
	(void)cmd_exec(cmdp, argp);
	exit(127);
	/*NOTREACHED*/

    default:	/* parent */
	*fdp = recv_pipe[0];
	(void)close(recv_pipe[1]);
	break;
    }
    return 0;
}

/*
 * cmd_start2: form a two-way pipe to the given command
 *   @cmdp:  command (may be simple or a /bin/sh command)
 *   @fdp:  pointer to read, write descriptor, respectively
 *   @pipd: process ID of child
 */
int
cmd_start2(const char *cmdp, int *fdp, pid_t *pidp)
{
    int recv_pipe[2];
    int send_pipe[2];

    if (pipe(recv_pipe) == -1) {
	  (void)fprintf(stderr, "%s: pipe: %s\n", progname, strerror(errno));
	  return -1;
    }
    if (pipe(send_pipe) == -1) {
	  (void)fprintf(stderr, "%s: pipe: %s\n", progname, strerror(errno));
	  return -1;
    }
    switch (*pidp = fork()) {
    case -1:
	(void)fprintf(stderr, "%s: fork: %s\n", progname, strerror(errno));
	return -1;

    case 0:	/* child */
	if (dup2(send_pipe[0], 0) == -1) {
	    (void)fprintf(stderr, "%s: dup2: %s\n", progname, strerror(errno));
	    exit(99);
	    /*NOTREACHED*/
	}
	if (dup2(recv_pipe[1], 1) == -1) {
	    (void)fprintf(stderr, "%s: dup2: %s\n", progname, strerror(errno));
	    exit(99);
	    /*NOTREACHED*/
	}
	(void)close(send_pipe[0]);
	(void)close(send_pipe[1]);
	(void)close(recv_pipe[0]);
	(void)close(recv_pipe[1]);
	(void)cmd_exec(cmdp, NULL);
	exit(127);
	/*NOTREACHED*/

    default:	/* parent */
	fdp[0] = recv_pipe[0];
	(void)close(recv_pipe[1]);
	(void)close(send_pipe[0]);
	fdp[1] = send_pipe[1];
	break;
    }
    return 0;
}
