/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <getopt.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "wfsyn_internal.h"
#include "wfsyn.h"
#include "command.h"
#include "calibrate_detector.h"
#include "calibrate_output.h"
#include "main.h"

char *progname;

static const char short_options[] = "+c:Dd:f:iPrv";

static const struct option long_options[] = {
    { "command",        1, NULL, 'c' },
    { "detector",	0, NULL, 'D' },
    { "device",		1, NULL, 'd' },
    { "frequency",      1, NULL, 'f' },
    { "interactive",    0, NULL, 'i' },
    { "power",   	0, NULL, 'P' },
    { "references",     0, NULL, 'r' },
    { "verbose",	0, NULL, 'v' },
    { NULL,             0, NULL,  0  }
};

static char const * const usage[] = {
    "-D|--detector [OPTIONS] [first last [step]]",
    "-D|--detector [OPTIONS] -r|--references reference-values...",
    "-P|--power    [OPTIONS] attenuation",
    NULL
};

static char const * const help[] = {
    "Operations:",
    "  -D, --detector           calibrate the power detector",
    "  -P, --power              calibrate the output power",
    "",
    "Common Options:",
    "  -d, --device=DEV         specify the Synth NV serial device",
    "  -v, --verbose            increase verbosity",
    "",
    "Power Detector Calibration",
    " Options:",
    "  -c, --command=CMD        specify a reference power control program",
    "  -i, --interactive        interact with the user or command",
    "  -r, --references         specify specific reference values in dBm",
    "  first last step          range of reference values in dBm",
    "  reference-values         specific reference values in dBm",
    "",
    "  Calibrates the Synth NV power detector using an external",
    "  reference source.  Reference power values may be specified as a",
    "  range or, if --references is given, a list of specific values.",
    "  If the --command option is given without --interactive, the given",
    "  program is executed once for each reference value with the value",
    "  passed on the command line; if the --command option is given",
    "  with --interactive, the program is executed once with reference",
    "  values given on stdin -- it should respond to each value by printing",
    "  \"ok\\n\" or a corrected power value on stdout when the reference",
    "  is settled.",
    "",
    "Output Power Calibration",
    " Options:",
    "  -f, --frequency=MHz      set the frequency to MHz.",
    "",
    "  Calibrates the Synth NV output power using the Synth NV power",
    "  detector.  The power detector should be calibrated first.",
    "  The Synth NV output must be connected to the power detector",
    "  through an optional attenuator.  The attenuator value must be",
    "  given on the command line.  A 20dB attenuator is recommended.",
    NULL
};
const char   *opt_c = NULL;
const char   *opt_d = "/dev/ttyACM0";
const char   *opt_f = NULL;
int8_t        opt_i = 0;
int8_t        opt_r = 0;
int           opt_v = 0;

/*
 * print_options
 */
void print_usage()
{
    char const * const * cpp;

    for (cpp = usage; *cpp != NULL; ++cpp) {
	(void)fprintf(stderr, "%s: usage %s\n", progname, *cpp);
    }
    for (cpp = help; *cpp != NULL; ++cpp) {
	(void)fprintf(stderr, "%s\n", *cpp);
    }
}

/*
 * error_fn: error function for wfsyn_openx
 *   Does not return.
 */
static void
error_fn(wfsyn_device_type *wfsynp, const char *message)
/*ARGSUSED*/
{
    (void)fprintf(stderr, "%s: %s\n", progname, message);
    exit(3);
    /*NOTREACHED*/
}

/*
 * main
 */
int
main(int argc, char **argv)
{
    int before_optind;
    char operation = '\000';
    wfsyn_device_type *wfsynp;
    char *cp = NULL;
    int fds[2];
    pid_t pid;
    double power0;	/* original output power setting */
    double s2 = 0.0;	/* sum of squares from means */

    /*
     * Parse options.
     */
    if ((char *)NULL == (progname = strrchr(argv[0], '/'))) {
	    progname = argv[0];
    } else {
	    ++progname;
    }
    opterr = 0;
    for (;;) {
	before_optind = optind;
	switch (getopt_long(argc, argv, short_options, long_options, NULL)) {
	case -1:
	    break;

	case 'c':
	    opt_c = optarg;
	    continue;

	case 'D':
	    if (operation != '\000' && operation != 'D') {
		(void)fprintf(stderr, "%s: -D is incompatible with -%c\n",
		              progname, operation);
		print_usage();
		exit(1);
		/*NOTREACHED*/
	    }
	    operation = 'D';
	    continue;

	case 'd':
	    opt_d = optarg;
	    continue;

	case 'f':
	    opt_f = optarg;
	    continue;

	case 'i':
	    opt_i = 1;
	    continue;

	case 'P':
	    if (operation != '\000' && operation != 'P') {
		(void)fprintf(stderr, "%s: -P is incompatible with -%c\n",
				progname, operation);
		print_usage();
		exit(1);
		/*NOTREACHED*/
	    }
	    operation = 'P';
	    continue;

	case 'r':
	    opt_r = 1;
	    continue;

	case 'v':
	    ++opt_v;
	    continue;

	default:
	    /*
	     * If an illegal option is a valid (negative) floating point
	     * number, then stop option parsing.  This allows negative
	     * numbers to appear as arguments without requiring a '--'
	     * separator.
	     */
	    (void)strtod(argv[before_optind], &cp);
	    if (cp > argv[before_optind] && *cp == '\000') {
		    break;
	    }
	    (void)fprintf(stderr, "%s: -%c: invalid option\n",
	    		  progname, optopt);
	    print_usage();
	    exit(1);
	    /*NOTREACHED*/
	}
	break;
    }
    argc -= optind;
    argv += optind;

    /*
     * If verbose, print the command and options.
     */
    if (opt_v > 0) {
	int i;

	(void)printf("# %s -%c -d %s -v", progname, operation, opt_d);
	if (opt_v > 1) {
	    (void)printf("v");
	    if (opt_v > 2) {
		(void)printf("v");
	    }
	}
	if (opt_c != NULL) {
	    (void)printf(" -c '%s'", opt_c);
	}
	if (opt_f != NULL) {
	    (void)printf(" -f %s", opt_f);
	}
	if (opt_i) {
	    (void)printf(" -i");
	}
	if (opt_r) {
	    (void)printf(" -r");
	}
	for (i = 0; i < argc; ++i) {
	    (void)printf(" %s", argv[i]);
	}
	(void)printf("\n");
    }

    /*
     * Handle operation-specific options and arguments.
     */
    switch (operation) {
    case 'D':
	 if (opt_f != NULL) {
	    (void)fprintf(stderr, "%s: -f is incompatible with -D\n", progname);
	    print_usage();
	    exit(1);
	    /*NOTREACHED*/
	 }
	 break;

    case 'P':
	if (opt_i) {
	    (void)fprintf(stderr, "%s: -i is incompatible with -P\n", progname);
	    print_usage();
	    exit(1);
	    /*NOTREACHED*/
	}
	if (opt_r) {
	    (void)fprintf(stderr, "%s: -r is incompatible with -P\n", progname);
	    print_usage();
	    exit(1);
	    /*NOTREACHED*/
	}
	if (opt_c != NULL) {
	    (void)fprintf(stderr, "%s: -s is incompatible with -P\n", progname);
	    print_usage();
	    exit(1);
	    /*NOTREACHED*/
	}
	break;

    default:
	(void)fprintf(stderr, "%s: no operation given\n", progname);
	print_usage();
	exit(1);
	/*NOTREACHED*/
    }
    if (!opt_c && !opt_i) {
	if (!isatty(0)) {
	    (void)fprintf(stderr, "%s: -S or -i required if no tty\n",
	    		  progname);
	    print_usage();
	    exit(1);
	    /*NOTREACHED*/
	}
	opt_i = 1;
    }

    /*
     * Close the send pipe and wait for the command to exit.
     */
    if (opt_c && opt_i) {
	if (cmd_start2(opt_c, fds, &pid) == -1) {
	    exit(2);
	    /*NOTREACHED*/
	}
    }

    /*
     * Open the wfsyn device.  On error, error_fn should exit.
     */
    wfsynp = wfsyn_openx(opt_d, WFSYN_OPENX_ERROR, error_fn);
    assert(wfsynp != NULL);

    /*
     * Remember the original output power then mute the output.
     */
    power0 = wfsyn_get_power(wfsynp);
    wfsyn_set_power(wfsynp, WFSYN_MUTE);

    /*
     * Do the specified operation.
     */
    switch (operation) {
    case 'D':
	{
	    double first = DEFAULT_DETECTOR_MIN;
	    double last  = DEFAULT_DETECTOR_MAX;
	    double step  = DEFAULT_DETECTOR_STEP;
	    int n_steps;
	    int n_measurements = 0;
	    int i;
	    measurement_type *mp;

	    if (!opt_r) {
		switch (argc) {
		case 3:
		    step = atof(argv[2]);
		    if (-EPS <= step && step <= EPS) {
			(void)fprintf(stderr, "%s: step size too small\n",
				      progname);
			exit(2);
			/*NOTREACHED*/
		    }
		    /*FALLTHROUGH*/

		case 2:
		    first = atof(argv[0]);
		    last  = atof(argv[1]);
		    /*FALLTHROUGH*/

		case 0:
		    break;

		default:
		    print_usage();
		    exit(1);
		    /*NOTREACHED*/
		}
		n_steps = 1 + (int)rint((last - first) / step);

	    } else {
		n_steps = argc;
	    }
	    if (n_steps < 2) {
		(void)fprintf(stderr, "%s: error: at least two reference "
			      "points are required\n", progname);
		print_usage();
		exit(1);
		/*NOTREACHED*/
	    }
	    mp = malloc(n_steps * sizeof(measurement_type));
	    if (mp == NULL) {
		(void)fprintf(stderr, "%s: malloc: %s\n",
			      progname, strerror(errno));
		exit(2);
		/*NOTREACHED*/
	    }
	    if (opt_v >= 3) {
		(void)printf("# raw power measurements\n");
	    }
	    for (i = 0; i < n_steps; ++i) {
		double y = opt_r ? atof(argv[i]) : first + step * i;
		response_type r;

		if (opt_c != NULL) {
		    if (opt_i) {
			r = set_power_ci(&y, fds);
		    } else {
			r = set_power_c(&y, opt_c);
		    }
		} else {
		    r = set_power_i(&y);
		}
		if (r == RESP_SKIP) {
		    continue;
		}
		if (r != RESP_OK) {
		    exit(1);
		    /*NOTREACHED*/
		}
		measure_power(wfsynp, y, &mp[n_measurements++], &s2);
	    }
	    if (n_measurements < 2) {
		(void)fprintf(stderr, "%s: error: at least two reference "
			      "points are required\n", progname);
		exit(2);
		/*NOTREACHED*/
	    }
	    if (opt_v >= 3) {
		(void)printf("\n\n");
	    }
	    calibrate_detector(wfsynp, mp, n_measurements,
	    		       N_SIGMA * sqrt(s2 / n_measurements));
	}
	break;

    case 'P':
	{
	    double attenuation;

	    if (argc != 1) {
		print_usage();
		exit(1);
		/*NOTREACHED*/
	    }
	    attenuation = atof(argv[0]);
	    if (attenuation < 0.0) {
		(void)fprintf(stderr, "%s: error: the attenuation value "
		              "must be specified as a positive number\n",
			      progname);
		exit(1);
		/*NOTREACHED*/
	    }
	    if (opt_f != NULL) {
		(void)wfsyn_set_frequency(wfsynp, atof(opt_f));
	    }
	    calibrate_output(wfsynp, attenuation);
	}
	break;
    }

    /*
     * Restore the original output power.
     */
    wfsyn_set_power(wfsynp, power0);

    /*
     * Save the new settings.
     */
    (void)wfsyn_save_settings(wfsynp);

    /*
     * Close the Synth NV device.
     */
    wfsyn_close(wfsynp);

    /*
     * Close the pipes and wait for the command to exit.
     */
    if (opt_c && opt_i) {
	(void)close(fds[0]);
	(void)close(fds[1]);
	(void)cmd_wait(pid);
    }

    exit(0);
    /*NOTREACHED*/
}
