/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

%{
#include <stdlib.h>
#include <stdint.h>
#include "wfsyn.h"
#include "error.h"
#include "lex.yy.h"
#include "main.h"

static wfsyn_options_type *woptp = NULL;

static int continuous_option;
static int sweep_step_type;
static char *sweep_calibration_filename;
static double sweep_hold_time;
static double icp_table[];

static void prompt();
static void parse_sweep_calibration(double *vector, int nsteps);

static void yyerror(const char *str);

%}

%token T_TRUE T_FALSE T_INT T_FLOAT T_STRING
       T_MUTE T_INTERNAL T_EXTERNAL T_FRAC_N T_LOW_SPUR T_INT_N
       T_MINUS T_PLUS T_ASSIGN
       T_CALIBRATION T_HOLD_TIME T_CONTINUOUS T_NO_CONTINUOUS T_LOG
       T_FREQUENCY T_POWER T_SHOW T_SAVE_SETTINGS T_SAVE_TO_NVRAM
       T_AM_BURST T_PULSE_BURST T_SWEEP T_DETECT_POWER
       T_READ_ANALOG1 T_READ_ANALOG2 T_DIGITAL1 T_DIGITAL2
       T_HELP T_SEMI T_NL T_EXIT

%nonassoc BELOW_OPTIONS

%left T_FMAX T_FMIN T_FREF T_FRES T_ICP T_MODE T_PFDMAX T_PFDMIN
      T_REFSRC T_RMUL2EN T_WLOCK

%union {
    double lf;
    int i;
    char *cp;
    wfsyn_mode_type mode;
    wfsyn_refsrc_type refsrc;
    wfsyn_icp_type icp;
}

%type <lf> float power
%type <i> int bool
%type <cp> string;
%type <mode> mode
%type <refsrc> refsrc
%type <icp> icp;

%start statement_list_opt

%%

statement_list_opt	: /*empty*/
			{
			    prompt();
			}
			| statement_list_opt statement
			;

statement		: T_FREQUENCY assign_opt float
			{
			    (void)wfsyn_set_frequency(wfsynp, $3);
			}
			| T_POWER assign_opt power
			{
			    (void)wfsyn_set_power(wfsynp, $3);
			}
			| {
			    woptp = wfsyn_get_options(wfsynp);
			} option_list {
			    (void)wfsyn_set_options(wfsynp);
			} %prec BELOW_OPTIONS
			| T_SHOW
			{
			    wfsyn_bool_type rmul2, rdiv2;
			    uint16_t r, mod;
			    const char *mode_name;

			    woptp = wfsyn_get_options(wfsynp);
			    (void)printf("model       \"%s\"\n",
					 wfsyn_get_model(wfsynp));
			    if (wfsyn_get_serial(wfsynp) != NULL) {
				(void)printf("serial      \"%s\"\n",
					     wfsyn_get_serial(wfsynp));
			    }
			    (void)printf("version     \"%s\"\n",
					 wfsyn_get_version(wfsynp));
			    (void)printf("\n");
			    (void)printf("frequency   %11.6f MHz\n",
			    		 wfsyn_get_frequency(wfsynp));
			    if (wfsyn_get_power(wfsynp) != WFSYN_MUTE) {
				(void)printf("power        %5.1f dBm\n",
					     wfsyn_get_power(wfsynp));
			    } else {
				(void)printf("power        MUTE\n");
				     
			    }
			    (void)printf("\n");
			    (void)printf("fmin        %11.6f MHz\n",
			    		 woptp->wopt_fmin);
			    (void)printf("fmax        %11.6f MHz\n",
			    		 woptp->wopt_fmax);
			    (void)printf("fref        %11.6f MHz\n",
			    		 woptp->wopt_fref);
			    (void)printf("fres        %11.6f MHz\n",
			    		 woptp->wopt_fres);
			    (void)printf("icp            %4.2f mA\n",
			    		 icp_table[(int)woptp->wopt_icp]);
			    switch (woptp->wopt_mode) {
			    case WFSYN_MODE_FRAC_N:
				mode_name = "FRAC_N";
				break;
			    case WFSYN_MODE_LOW_SPUR:
				mode_name = "LOW_SPUR";
				break;
			    case WFSYN_MODE_INT_N:
				mode_name = "INT_N";
				break;
			    default:
				abort();
			    }
			    (void)printf("mode        %s\n", mode_name);
			    (void)printf("pfdmin      %11.6f MHz\n",
			    		 woptp->wopt_pfdmin);
			    (void)printf("pfdmax      %11.6f MHz\n",
			    		 woptp->wopt_pfdmax);
			    (void)printf("refsrc         %d\n",
			    		 woptp->wopt_refsrc);
			    (void)printf("rmul2en        %d\n",
			    		 woptp->wopt_rmul2en);
			    (void)printf("wlock          %d\n",
			    		 woptp->wopt_wlock);
			    (void)printf("\n");
			    (void)printf("PFD         %11.6f MHz\n",
			                 wfsyn_get_pfd(wfsynp));
			    (void)wfsyn_get_rmod(wfsynp, &rmul2, &rdiv2,
			    			 &r, &mod);
			    (void)printf("RMUL2          %d\n", rmul2);
			    (void)printf("RDIV2          %d\n", rdiv2);
			    (void)printf("R           %4u\n",  r);
			    (void)printf("MOD         %4u\n", mod);
			    (void)fflush(stdout);
			}
			| T_SAVE_SETTINGS
			{
			    (void)wfsyn_save_settings(wfsynp);
			}
			| T_SAVE_TO_NVRAM
			{
			    (void)wfsyn_save_to_nvram(wfsynp);
			}
			| T_AM_BURST burst_option_list_opt float float int
			{
			    if (continuous_option == 0) {
				wfsyn_am_continuous(wfsynp, 0);
			    }
			    (void)wfsyn_am_burst(wfsynp, $3, $4, $5);
			    if (continuous_option == 1) {
				wfsyn_am_continuous(wfsynp, 1);
			    }
			}
			| T_PULSE_BURST burst_option_list_opt float float int
			{
			    if (continuous_option == 0) {
				wfsyn_am_continuous(wfsynp, 0);
			    }
			    (void)wfsyn_pulse_burst(wfsynp, $3, $4, $5);
			    if (continuous_option == 1) {
				wfsyn_am_continuous(wfsynp, 1);
			    }
			}
			| T_SWEEP sweep_option_list_opt float float float
			{
			    int i;
			    int n = 0;
			    int nsteps;
			    double fmin = $3;
			    double fmax = $4;
			    double *calibration_vector = NULL;
			    double *frequency_vector;
			    double *result_vector;

			    /*
			     * Find the number of steps then adjust fmax
			     * to respect the requested step size.
			     */
			    if (sweep_step_type == WFSYN_STEP_LOG) {
				if ($3 > 0.0 && $4 > 0.0 && $5 > 0.0) {
				    n = (int)rint(log($4 / $3) / log($5 / $3));
				    fmax = fmin * pow($5 / $3, (double)n);
				}
			    } else {
				if ($4 >= $3 && $5 > 0) {
				    n = (int)rint(($4 - $3) / $5);
				    fmax = fmin + n * $5;
				}
			    }
			    nsteps = n + 1;

			    /*
			     * If a calibration file was given, allocate
			     * the vector and parse the calibration file.
			     */
			    if (sweep_calibration_filename != NULL) {
				calibration_vector = malloc(nsteps *
							    sizeof(double));
				if (calibration_vector == NULL) {
				    error(ERR_FATAL, "malloc: %s",
				          strerror(errno));
				}
				parse_sweep_calibration(calibration_vector,
						        nsteps);
				free((void *)sweep_calibration_filename);
				sweep_calibration_filename = NULL;
			    }

			    /*
			     * Allocate the frequency and dBm vectors.
			     */
			    if ((frequency_vector = malloc(nsteps *
					    sizeof(double))) == NULL) {
				error(ERR_FATAL, "malloc: %s", strerror(errno));
			    }
			    if ((result_vector = malloc(nsteps *
					    sizeof(double))) == NULL) {
				error(ERR_FATAL, "malloc: %s", strerror(errno));
			    }

			    /*
			     * Run the sweep.
			     */
			    (void)wfsyn_sweep(wfsynp, fmin, fmax, nsteps,
			    		      sweep_step_type, sweep_hold_time,
					      calibration_vector,
					      frequency_vector, result_vector);
			    for (i = 0; i < nsteps; ++i) {
				(void)printf("%11.6f %6.2f\n",
					     frequency_vector[i],
					     result_vector[i]);
			    }
			    (void)fflush(stdout);
			    free((void *)result_vector);
			    free((void *)frequency_vector);
			    if (calibration_vector != NULL) {
				free((void *)calibration_vector);
			    }
			}
			| T_DETECT_POWER
			{
			    (void)printf("%5.1f\n", wfsyn_detect_power(wfsynp));
			    (void)fflush(stdout);
			}
			| T_READ_ANALOG1
			{
			    (void)printf("%5.3f\n", wfsyn_read_analog1(wfsynp));
			}
			| T_READ_ANALOG2
			{
			    (void)printf("%5.3f\n", wfsyn_read_analog2(wfsynp));
			}
			| T_DIGITAL1 assign_opt bool
			{
			    (void)wfsyn_write_digital1(wfsynp, $3);
			}
			| T_DIGITAL2 assign_opt bool
			{
			    (void)wfsyn_write_digital2(wfsynp, $3);
			}
			| T_HELP
			{
			    print_usage();
			}
			| T_EXIT
			{
			    exit(0);
			    /*NOTREACHED*/
			}
			| T_STRING
			{
			    error(ERR_ERROR, "unrecognized keyword `%s'",
			          yytext);
			}
			| T_SEMI
			{
			    /*NULL*/;
			}
			| T_NL
			{
			    prompt();
			}
			| error T_NL
			{
			    prompt();
			}
			;

assign_opt		: /*empty*/
			| T_ASSIGN
			;

power			: float
			{
				$$ = $1;
			}
			| T_MUTE
			{
				$$ = WFSYN_MUTE;
			}
			;

option_list		: option
			| option_list option
			;

option			: T_FREF assign_opt float
			{
			    woptp->wopt_fref = $3;
			}
			| T_FRES assign_opt float
			{
			    woptp->wopt_fres = $3;
			}
			| T_FMIN assign_opt float
			{
			    woptp->wopt_fmin = $3;
			}
			| T_FMAX assign_opt float
			{
			    woptp->wopt_fmax = $3;
			}
			| T_PFDMIN assign_opt float
			{
			    woptp->wopt_pfdmin = $3;
			}
			| T_PFDMAX assign_opt float
			{
			    woptp->wopt_pfdmax = $3;
			}
			| T_MODE assign_opt mode
			{
			    woptp->wopt_mode = $3;
			}
			| T_REFSRC assign_opt refsrc
			{
			    woptp->wopt_refsrc = $3;
			}
			| T_RMUL2EN assign_opt bool
			{
			    woptp->wopt_rmul2en = $3;
			}
			| T_WLOCK assign_opt bool
			{
			    woptp->wopt_wlock = $3;
			}
			| T_ICP assign_opt icp
			{
			    woptp->wopt_icp = $3;
			}
			;

burst_option_list_opt	: /*empty*/
			{
			    /* set defaults for all options */
			    continuous_option = -1;
			}
			| burst_option_list_opt T_CONTINUOUS
			{
			    continuous_option =  1;
			}
			| burst_option_list_opt T_NO_CONTINUOUS
			{
			    continuous_option =  0;
			}
			;

sweep_option_list_opt	: /*empty*/
			{
			    /* set defaults for all options */
			    sweep_calibration_filename = NULL;
			    sweep_hold_time = 0.0;
			    sweep_step_type = WFSYN_STEP_LIN;
			}
			| sweep_option_list_opt T_CALIBRATION assign_opt string
			{
			    if (sweep_calibration_filename != NULL) {
				error(ERR_WARNING, "multiple --calibration "
				                   "options given");
				free((void *)sweep_calibration_filename);
			    }
			    sweep_calibration_filename = $4;
			}
			| sweep_option_list_opt T_HOLD_TIME assign_opt float
			{
			    sweep_hold_time = $4;
			}
			| sweep_option_list_opt T_LOG
			{
			    sweep_step_type = WFSYN_STEP_LOG;
			}
			;

string			: T_STRING
			{
			    if (($$ = strdup(yytext)) == NULL) {
				error(ERR_FATAL, "malloc: %s", strerror(errno));
			    }
			}
			;

float			: T_FLOAT
			{
			    $$ = atof(yytext);
			}
			| int
			{
			    $$ = (double)$1;
			}
			| T_MINUS float
			{
			    $$ = -$2;
			}
			| T_PLUS float
			{
			    $$ = $2;
			}
			;

int			: T_INT
			{
			    $$ = atoi(yytext);
			}
			;

bool			: T_TRUE
			{
			    $$ = 1;
			}
			| T_FALSE
			{
			    $$ = 0;
			}
			| int
			{
			    $$ = ($1 != 0);
			}
			;

mode			: T_FRAC_N
			{
			    $$ = WFSYN_MODE_FRAC_N;
			}
			| T_LOW_SPUR
			{
			    $$ = WFSYN_MODE_LOW_SPUR;
			}
			| T_INT_N
			{
			    $$ = WFSYN_MODE_INT_N;
			}
			;

refsrc			: T_EXTERNAL
			{
			    $$ = WFSYN_REFSRC_EXTERNAL;
			}
			| T_INTERNAL
			{
			    $$ = WFSYN_REFSRC_INTERNAL;
			}
			;

icp			: float
			{
			    int i;
			    int best_i = -1;
			    double best_err;

			    for (i = 15; i >= 0; --i) {
				double err = fabs($1 - icp_table[i]);

				if (best_i == -1 || err < best_err) {
				    best_i = i;
				    best_err = err;
				}
			    }
			    return (wfsyn_icp_type)i;
			}
			;


%%

static double icp_table[] = {
    0.31, 0.63, 0.94, 1.25, 1.56, 1.88, 2.19, 2.50,
    2.81, 3.13, 3.44, 3.75, 4.06, 4.38, 4.69, 5.00
};

/*
 * prompt
 */
static void
prompt()
{
    if (interactive) {
	(void)printf("> ");
	(void)fflush(stdout);
    }
}

#define MAX_LINE	80

/*
 * parse_sweep_calibration
 */
static void parse_sweep_calibration(double *vector, int nsteps)
{
    FILE *fp;
    char buffer[MAX_LINE+2];
    int lineno = 0;
    int count = 0;
    int ncols = -1;

    (void)memset((void *)vector, 0, nsteps * sizeof(double));
    fp = fopen(sweep_calibration_filename, "r");
    if (fp == NULL) {
	error(ERR_ERROR, "%s: fopen: %s",
	      sweep_calibration_filename, strerror(errno));
	return;
    }
    while (fgets(buffer, MAX_LINE + 2, fp) != NULL) {
	char *cp;
	double lf1, lf2;
	int n;

	/*
	 * Force NUL-termination and count the line.
	 */
	buffer[MAX_LINE] = '\000';
	++lineno;

	/*
	 * Remove comments and the terminating newline.
	 */
	if ((cp = strpbrk(buffer, "#\n")) == NULL) {
	    error(ERR_ERROR, "%s: line %d: line too long\n",
	          sweep_calibration_filename, lineno);
	    (void)fclose(fp);
	    return;
	}
	*cp = '\000';

	/*
	 * Skip leading whitespace.
	 */
	cp = buffer;
	while (strchr(" \t\r\f\v", *cp) != NULL) {
	    ++cp;
	}

	/*
	 * Ignore blank lines.
	 */
	if (*cp == '\000') {
	    continue;
	}

	/*
	 * Scan values from the line.
	 */
	n = sscanf(cp, " %lf %lf", &lf1, &lf2);
	if (n <= 0) {
	    error(ERR_ERROR, "%s: line %d: bad data\n",
		  sweep_calibration_filename, lineno);
	    (void)fclose(fp);
	    return;
	}

	/*
	 * Check for too many values.
	 */
	if (n > 0 && count == nsteps) {
	    error(ERR_ERROR, "%s: too many calibration values (expected %d)",
		  sweep_calibration_filename, nsteps);
	    (void)fclose(fp);
	    return;
	}

	/*
	 * On the first data line, note the number of columns scanned
	 * (up to 2).  On subsequent lines, make sure the number is
	 * the same.
	 */
	if (ncols == -1) {
	    ncols = n;
	} else if (n != ncols) {
	    error(ERR_ERROR, "%s: line %d: inconsistent number of data "
	    		     "columns d (%d != %d)", 
		  sweep_calibration_filename, lineno, n, ncols);
	    (void)fclose(fp);
	    return;
	}

	/*
	 * If there are at least two columns, take the second column;
	 * if there is only one column, take the first column.
	 */
	vector[count++] = (ncols == 1) ? lf1 : lf2;
    }
    /*
     * Check for too few values.
     */
    if (count != nsteps) {
	error(ERR_ERROR, "%s: too few calibration values "
	                 "(expected %d, found %d)",
	      sweep_calibration_filename, nsteps, count);
	(void)fclose(fp);
	return;
    }
    (void)fclose(fp);
}

/*
 * yyerror
 */
void
yyerror(const char *str)
{
    error(ERR_ERROR, str);
}
