/*
 * Windfreak Technologies SynthNV Series RF Signal Generator Library
 * Copyright © 2013 D Scott Guthridge <pdx_scooter@yahoo.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <assert.h>
#include <errno.h>
#include <getopt.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "wfsyn.h"
#include "lex.yy.h"
#include "error.h"
#include "main.h"

char *progname;

/*
 * True if reading from a terminal.
 */
int interactive;

/*
 * Symbols from scan.l and parse.y
 */
extern int (*yyinput_fn)(char *buffer, int count);
extern int arg_yyin(char *buffer, int count);
extern int file_yyin(char *buffer, int count);
extern int yyparse(void);

wfsyn_device_type *wfsynp = NULL;


/*
 * wfsyn_err: report errors from the wfsyn library
 */
static void wfsyn_err(const wfsyn_device_type *wfsynp, const char *msg)
{
    error(ERR_ERROR, msg);
}

static const char short_options[] = "+d:e:f:";

static const struct option long_options[] = {
    { "device",		1, NULL, 'd' },
    { "expression",	1, NULL, 'e' },
    { "file",		1, NULL, 'f' },
    { NULL,             0, NULL,  0  }
};

static char const * const usage[] = {
    "[-d|--device=DEV] [-e|--expression=CMDS] [-f|--file=FILE] [CMDS...]",
    NULL
};

static char const * const help[] = {
    "-d, --device=DEV              specify the Synth NV serial device",
    "-e, --expression=CMDS         process commands in CMDS",
    "-f, --file=FILE               read commands from file (- for stdin)",
    "CMDS...                       process commands in CMDS",
    "",
    "Command Syntax:",
    "  f|frequency = MHz           set the synthesizer frequency",
    "  p|power     = dBm           set the output power",
    "  option=value option=value option=value ... \\",
    "               option=value option=value ...  set options",
    "    options are:",
    "      fref        = MHz                      reference frequency",
    "      fres        = MHz                      frequency resolution",
    "      fmin        = MHz                      minimum output frequency",
    "      fmax        = MHz                      maximum output frequency",
    "      pfdmin      = MHz                      minimum phase det frequency",
    "      pfdmax      = MHz                      maximum phase det frequency",
    "      mode        = {FRAC_N|LOW_SPUR|INT_N}  synthesizer mode",
    "      refsrc      = {internal|external}      reference source",
    "      rmul2en     = {true|false}             enable reference doubler",
    "      wlock       = {true|false}             wait for phase lock",
    "      icp         = mA                       charge pump current",
    "",
    "  show                        show current settings",
    "",
    "  save-settings               save current settings to the start-up file",
    "",
    "  save-to-nvram               save current settings to NVRAM",
    "",
    "  am-burst [-c|--continuous|-C|--no-continuous] mod-freq mod-depth nreps",
    "    -c, --continuous          enable continous AM bursts",
    "    -C, --no-continuous       disable continuous AM bursts",
    "    mod-freq                  modulation frequency (Hz)",
    "    mod-depth                 modulation depth (percent)",
    "    nreps                     number of modulation cycles per burst",
    "",
    "  pulse-burst [-c|--continuous|-C|--no-continuous] on-time off-time nreps",
    "    -c, --continuous          enable continous pulse bursts",
    "    -C, --no-continuous       disable continuous pulse bursts",
    "    on-time                   pulse on-time  in us",
    "    off-time                  pulse off-time in us",
    "    nreps                     number of pulses per burst",
    "",
    "  sweep [--calibration file] [--hold-time time] [--log] fmin fmax fstep",
    "    --calibration file        use previous sweep output as baseline",
    "    --hold-time time          hold each frequency for time us",
    "    --log                     space steps equally on a log scale",
    "    fmin, fmax                minimum and maximum frequency",
    "    fstep                     size of first step",
    "",
    "  d|detect-power              read the power detector",
    "  a1|read-analog1             read the analog voltage at P8, pin 1",
    "  a2|read-analog2             read the analog voltage at P8, pin 2",
    "  d1|digital1 = {true|false}  set the P8, pin 3 digital output",
    "  d1|digital2 = {true|false}  set the P8, pin 5 digital output",
    "",
    "  help                        show this help message",
    "",
    "  exit|quit                   exit the program",
    "",
    NULL
};
static char *device_name = "/dev/ttyACM0";

/*
 * print_options
 */
void print_usage()
{
    char const * const * cpp;

    for (cpp = usage; *cpp != NULL; ++cpp) {
	(void)fprintf(stderr, "%s: usage %s\n", progname, *cpp);
    }
    for (cpp = help; *cpp != NULL; ++cpp) {
	(void)fprintf(stderr, "%s\n", *cpp);
    }
}

/*
 * globals
 */
const char *input_filename = "<cmdline>";
int parse_argc;
char **parse_argv;
int parse_offset;

/*
 * open_device: open the Synth NV device if not alreaday open
 */
static void open_device()
{
    if (wfsynp == NULL) {
	wfsynp = wfsyn_openx(device_name, WFSYN_OPENX_ERROR, wfsyn_err);
	if (wfsynp == NULL) {
	    exit(2);
	    /*NOTREACHED*/
	}
    }
}

/*
 * parse_arguments: parse argc arguments given in argv
 */
static void parse_arguments(int argc, char **argv)
{
    input_filename = "<cmdline>";
    parse_argc = argc;
    parse_argv = argv;
    parse_offset = 0;
    interactive = 0;
    yyinput_fn = arg_yyin;
    yyset_lineno(1);
    open_device();
    yyparse();
}

/*
 * parse_file: parse the given file
 */
static void parse_file(const char *filename)
{
    input_filename = filename;
    if (filename[0] == '-' && filename[1] == '\000') {
	input_filename = "<stdin>";
	clearerr(stdin);
	yyin = stdin;
    } else if ((yyin = fopen(optarg, "r")) == NULL) {
	error(ERR_FATAL, "%s", strerror(errno));
	/*NOTREACHED*/
    }
    interactive = isatty(fileno(yyin));
    yyinput_fn = file_yyin;
    yyset_lineno(1);
    open_device();
    yyparse();
    if (yyin != stdin) {
	(void)fclose(yyin);
    }
    yyin = NULL;
}

/*
 * main
 */
int
main(int argc, char **argv)
{
    if ((char *)NULL == (progname = strrchr(argv[0], '/'))) {
	    progname = argv[0];
    } else {
	    ++progname;
    }
    yyin = NULL;
    for (;;) {
	switch (getopt_long(argc, argv, short_options, long_options, NULL)) {
	case -1:
	    break;

	case 'd':
	    if (wfsynp != NULL) {
		wfsyn_close(wfsynp);
		wfsynp = NULL;
	    }
	    device_name = optarg;
	    open_device();
	    continue;

	case 'e':
	    parse_arguments(1, &optarg);
	    continue;

	case 'f':
	    parse_file(optarg);
	    continue;

	default:
	    print_usage();
	    exit(1);
	    /*NOTREACHED*/
	}
	break;
    }

    /*
     * Parse argument options.  If none are given, parse stdin.
     */
    argc -= optind;
    argv += optind;
    if (argc != 0) {
	parse_arguments(argc, argv);
    } else {
	parse_file("-");
    }
    wfsyn_close(wfsynp);
    exit(0);
    /*NOTREACHED*/
}
